# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import os
from appdirs import user_cache_dir
import tempfile

from ..commons import logger

def assert_folder(folder):
    # make sure the folder exists and is writable, then return it.
    # If it cannot be written, create and return
    # a temporary folder.
    try:
        if not os.path.isdir(folder):
            os.makedirs(folder)
        if not os.access(folder, os.W_OK):
            raise OSError
        return folder
    except OSError:
        # cannot write to requested directory, create a temporary one.
        tmpdir = os.environ["SIIBRA_CACHEDIR"] = \
            tempfile.mkdtemp(prefix="siibra-cache-")
        logger.warning(
            f"Siibra created a temporary cache directory at {tmpdir}, as "
            f"the requested folder ({folder}) was not usable. "
            "Please consider to set the SIIBRA_CACHEDIR environment variable "
            "to a suitable directory.")
        return tmpdir


class Cache:

    _instance = None
    folder = user_cache_dir(".".join(__name__.split(".")[:-1]), "")

    def __init__(self):
        raise RuntimeError(
            "Call instance() to access "
            f"{self.__class__.__name__}")

    @classmethod
    def instance(cls):
        """
        Return an instance of the siibra cache. Create folder if needed.
        """
        if cls._instance is None:
            if "SIIBRA_CACHEDIR" in os.environ:
                cls.folder = os.environ["SIIBRA_CACHEDIR"]
            cls.folder = assert_folder(cls.folder)
            cls._instance = cls.__new__(cls)
        return cls._instance

    def clear(self):
        import shutil

        logger.info(f"Clearing siibra cache at {self.folder}")
        shutil.rmtree(self.folder)
        self.folder = assert_folder(self.folder)

    def build_filename(self, str_rep, suffix=None):
        hashfile = os.path.join(
            self.folder,
            str(hashlib.sha256(str_rep.encode("ascii")).hexdigest())
        )
        if suffix is None:
            return hashfile
        else:
            return hashfile + "." + suffix


CACHE = Cache.instance()



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ctypes import ArgumentError
from ..commons import logger
from ..core.space import Space, Point, PointSet
from ..core.parcellation import Parcellation

import numpy as np
from skimage import measure
from nilearn import image
from tqdm import tqdm

class BigBrainCortexSampler:
    """ Sample layerwise grayvalue statistics from BigBrain cortex.

    Given a coordinate in BigBrain space, the sampler centers a fixed-size box 
    near the given location in BigBrain cortex and extracts layerwise grayvalue 
    statistics. The center is found by identifying the closest point in Layer IV
    which is inside a specified maximum distance (default 1mm) from the given point.
    """
    
    def __init__(self, boxwidth_mm=3, maxdist_mm=1 ):
        """ Create a new cortex sample for BigBrain.
        
        Arguments
        ---------
        boxwidth_mm : float
            Sidelength of the boxes used for grayvalue sampling (default: 3mm)
        maxdist_mm : float
            Maximum distance allowed to shift given sampling locations towards the 
            mid surface of the cortex. If the mid surface is not found within
            this distance, no data will be sampled.
        """
        
        self.space = Space.REGISTRY['bigbrain']
        self._layermask = Parcellation.REGISTRY['layers'].get_map(self.space)
        self._template = self.space.get_template()
        self.boxwidth_mm = boxwidth_mm
        self.maxdist_mm = maxdist_mm
        
    def sample(self, location):
        """
        Given sample locations in BigBrain space, the sampler centers a fixed-size box 
        near the given location in BigBrain cortex and extracts layerwise grayvalue 
        statistics. The center is found by identifying the closest point in Layer IV
        which is inside a specified maximum distance (default 1mm) from the given point.
        If the location is not close enough to the cortical midsurface, the data sample
        will be empty and a warning will be printed.

        Parameters
        ----------
        location: Point, or PointSet
            Candidate location(s) for sampling
        
        Return
        ------
        List of dicts, one per sample point, with keys:
            - 'center': the physical coordinate of the cube used as a region of interest
            - 'boxsize': sidelenght in mm f the cube used as a region of interest
            - 'space': name of the space (bigbrain)
            - 'layers': Dict of layer-wise statistics with mean gray value, standard deviation, and volume in mm
        """
        if location.space != self.space:
            logger.info(f"Warping sample locations from {location.space.name} to {self.space.name}")
            loc_bb = location.warp(self.space)
        else:
            loc_bb = location

        result = []
        if isinstance(loc_bb, Point):
            result.append(self._sample_single_point(loc_bb))
        elif isinstance(loc_bb, PointSet):
            for p in tqdm(
                loc_bb, total=len(loc_bb), unit="locations",
                desc=f"Sampling from {len(loc_bb)} locations",
                disable=logger.level>20
            ):
                result.append(self._sample_single_point(p))
        else:
            raise ArgumentError(f"Invalid location specification {location.__class__.__name__} BigBrain sampling.")

        return result


    def _sample_single_point(self, point):
        """
        Given a coordinate in BigBrain space, the sampler centers a fixed-size box 
        near the given location in BigBrain cortex and extracts layerwise grayvalue 
        statistics. The center is found by identifying the closest point in Layer IV
        which is inside a specified maximum distance (default 1mm) from the given point.

        Parameters
        ----------
        point: Point
            Candidate location for sampling

        Return
        ------
        dict with keys:
            - 'center': the physical coordinate of the cube used as a region of interest
            - 'boxsize': sidelenght in mm f the cube used as a region of interest
            - 'space': name of the space (bigbrain)
            - 'layers': Dict of layer-wise statistics with mean gray value, standard deviation, and volume in mm
        
        """    
        result = {
            'center': tuple(point), 
            'boxsize': self.boxwidth_mm, 
            'space': self.space.name,
            'layers': {}
        }

        # find closest point in layer 4, i.e. towards the mid surface,
        # to make sure the box is well centered in the cortex.
        voi = point.get_enclosing_cube(2*self.maxdist_mm+1)
        voimask = self._layermask.fetch(voi=voi, resolution_mm=-1)
        L = np.asanyarray(voimask.dataobj)
        XYZ_ = np.array(np.where(L==4)).T
        if XYZ_.shape[0]==0:
            logger.warn(
                f"The point {tuple(point)} seems too far away from Layer IV in BigBrain. "
                f"No data sampled."
            )
            return result
        XYZ = np.dot(voimask.affine, np.c_[XYZ_, np.ones(XYZ_.shape[0])].T)[:3,:].T
        D = np.sum(np.sqrt((XYZ - tuple(point))**2), axis=1)
        p_mid = Point(
            XYZ[np.argmin(D),:], 
            self.space
        )

        # Load cube of fixed size at this cortica position
        # from BigBrain and cortical layer maps,
        # using maximum available resolution.
        voi = p_mid.get_enclosing_cube(self.boxwidth_mm)
        voidata = self._template.fetch(voi=voi, resolution_mm=-1)
        voimask = image.resample_to_img(
            self._layermask.fetch(voi=voi, resolution_mm=-1),
            voidata,
            interpolation='nearest'
        )
        scale = np.linalg.det(voimask.affine)
        
        # Get layer mask with possible additional segmentation 
        # components at box borders suppressed (e.g. from neighboring 
        # sulcal walls)
        L = np.asanyarray(voimask.dataobj)
        M = measure.label(L>0)
        cx, cy, cz = np.array(M.shape)//2
        L[M!=M[cx, cy, cz]] = 0
         
        # extract layerwise grayvalues, excluding value 255
        # which we assume to belong to background. 
        result['center'] = tuple(p_mid)
        I = np.asanyarray(voidata.dataobj)
        for l in range(1,7):
            mask = (L==l) & (I!=255)
            values = I[mask].ravel()
            result['layers'][l] = {
                'mean': values.mean(), 
                'std': values.std(),
                'volume': np.count_nonzero(mask)*scale
            }
            
        return result 
    


# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .feature import RegionalFeature
from .query import FeatureQuery

from ..commons import logger
from ..core.space import Space, Point
from ..retrieval.repositories import GitlabConnector, OwncloudConnector

import numpy as np
import os
import importlib
import pandas as pd


class CorticalCellDistribution(RegionalFeature):
    """
    Represents a cortical cell distribution dataset.
    Implements lazy and cached loading of actual data.
    """

    def __init__(self, regionspec: str, cells: pd.DataFrame, connector, folder, species):

        _, section_id, patch_id = folder.split("/")
        RegionalFeature.__init__(self, regionspec, species=species)
        self.cells = cells
        self.section = section_id
        self.patch = patch_id

        # construct lazy data loaders
        self._info_loader = connector.get_loader(
            "info.txt", folder, CorticalCellDistribution._decode_infotxt
        )
        self._image_loader = connector.get_loader("image.nii.gz", folder)
        self._layerinfo_loader = connector.get_loader(
            "layerinfo.txt", folder, CorticalCellDistribution._decode_layerinfo
        )
        self._connector = connector

    @staticmethod
    def _decode_infotxt(b):
        return dict(_.split(" ") for _ in b.decode("utf8").strip().split("\n"))

    @staticmethod
    def _decode_layerinfo(b):
        data =  np.array(
            [tuple(_.split(" ")) for _ in b.decode().strip().split("\n")[1:]],
            dtype=[
                ("index", "i"),
                ("label", "U10"),
                ("area (micron^2)", "f"),
                ("avg. thickness (micron)", "f"),
            ],
        )
        return pd.DataFrame(data)

    def layer_density(self, layerindex: int):
        """ Compute density of segmented cells for given layer in cells / mm^2. """
        assert layerindex in range(1, 7)
        num_cells = self.cells[self.cells['layer']==layerindex].shape[0]
        area = self.layers[self.layers['index']==layerindex].iloc[0]['area (micron^2)'] 
        return num_cells / area * 1e3**2

    def load_segmentations(self):
        from PIL import Image
        from io import BytesIO

        def imgdecoder(b):
            return np.array(Image.open(BytesIO(b)))

        index = 0
        result = []
        logger.info(f"Loading cell instance segmentation masks for {self}...")
        while True:
            try:
                target = f"segments_cpn_{index:02d}.tif"

                loader = self._connector.get_loader(target, self.folder, imgdecoder)
                result.append(loader.data)
                index += 1
            except RuntimeError:
                break
        return result

    @property
    def info(self):
        return self._info_loader.data

    @property
    def layers(self):
        """
        6x4 array of cortical layer attributes:
            Number Name Area(micron**2) AvgThickness(micron)
        """
        return self._layerinfo_loader.data

    @property
    def image(self):
        """
        Nifti1Image representation of the original image patch,
        with an affine matching it to the histological BigBrain space.
        """
        return self._image_loader.data

    @property
    def location(self):
        """
        Location of this image patch in BigBrain histological space in mm.
        """
        return Point(
            np.dot(self.image.affine, [0, 0, 0, 1])[:3], 
            Space.REGISTRY.BIG_BRAIN,
        )

    def __str__(self):
        return f"BigBrain cortical cell distribution in {self.regionspec} (section {self.info['section_id']}, patch {self.info['patch_id']})"

    def plot(self, title=None):
        """
        Create & return a matplotlib figure illustrating the patch,
        detected cells, and location in BigBrain space.
        """
        for pkg in ["matplotlib", "nilearn"]:
            if importlib.util.find_spec(pkg) is None:
                logger.warning(f"{pkg} not available. Plotting disabled.")
                return None

        from matplotlib import pyplot
        from nilearn import plotting

        patch = self.image.get_fdata()
        tpl = self.location.space.get_template().fetch()
        fig = pyplot.figure(figsize=(12, 6))
        pyplot.suptitle(str(self))
        ax1 = pyplot.subplot2grid((1, 4), (0, 0))
        ax2 = pyplot.subplot2grid((1, 4), (0, 1), sharex=ax1, sharey=ax1)
        ax3 = pyplot.subplot2grid((1, 4), (0, 2), colspan=2)
        ax1.imshow(patch, cmap="gray")
        ax2.axis("off")
        ax2.imshow(patch, cmap="gray")
        ax2.scatter(
            self.cells['x'],
            self.cells['y'],
            s=np.sqrt(self.cells['area']),
            c=self.cells['layer'])
        ax2.axis("off")
        view = plotting.plot_img(
            tpl, cut_coords=tuple(self.location), cmap="gray", axes=ax3, display_mode="tiled"
        )
        view.add_markers([tuple(self.location)])
        return fig


class RegionalCellDensityExtractor(FeatureQuery):

    _FEATURETYPE = CorticalCellDistribution
    _JUGIT = GitlabConnector("https://jugit.fz-juelich.de", 4790, "v1.0a1")
    _SCIEBO = OwncloudConnector("https://fz-juelich.sciebo.de", "yDZfhxlXj6YW7KO")


    def __init__(self, **kwargs):
        FeatureQuery.__init__(self)
        logger.warning(
            f"PREVIEW DATA! {self._FEATURETYPE.__name__} data is only a pre-release snapshot. Contact support@ebrains.eu if you intend to use this data."
        )

        species = {
            '@id': 'https://nexus.humanbrainproject.org/v0/data/minds/core/species/v1.0.0/0ea4e6ba-2681-4f7d-9fa9-49b915caaac9', 
            'name': 'Homo sapiens'
        }
        for cellfile, loader in self._JUGIT.get_loaders(
            suffix="segments.txt", recursive=True, decode_func=lambda b: b.decode()
        ):
            region_folder = os.path.dirname(cellfile)
            regionspec = " ".join(region_folder.split(os.path.sep)[0].split("_")[1:])
            cells = np.array(
                [
                    tuple(float(w) for w in _.strip().split(" "))
                    for _ in loader.data.strip().split("\n")[1:]
                ],
                dtype=[
                    ("x", "f"),
                    ("y", "f"),
                    ("area", "f"),
                    ("layer", "<i8"),
                    ("instance label", "<i"),
                ]
            )
            self.register(
                CorticalCellDistribution(regionspec, pd.DataFrame(cells), self._SCIEBO, region_folder, species=species)
            )



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .commons import logger, QUIET, VERBOSE

# __version__ is parsed by setup.py
__version__ = "0.3a8"
logger.info(f"Version: {__version__}")
logger.warning("This is a development release. Use at your own risk.")
logger.info(
    "Please file bugs and issues at https://github.com/FZJ-INM1-BDA/siibra-python."
)

from .core import spaces, parcellations, atlases
from .features import modalities, gene_names, get_features
from .commons import MapType, ParcellationIndex
from .retrieval import EbrainsRequest, CACHE
from .core import Point, PointSet, BoundingBox
from .core.space import Location as _
from . import samplers
from_sands = _.from_sands
set_ebrains_token = EbrainsRequest.set_token
fetch_ebrains_token = EbrainsRequest.fetch_token
clear_cache = CACHE.clear


def set_feasible_download_size(maxsize_gbyte):
    from .volumes import volume
    volume.gbyte_feasible = maxsize_gbyte
    logger.info(f"Set feasible download size to {maxsize_gbyte} GByte.")



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..commons import logger
from ..retrieval import EbrainsKgQuery

import re


class Dataset:
    """Parent class for datasets. Each dataset has an identifier."""

    REGISTRY = {}

    def __init__(self, identifier, description=""):
        self.id = identifier
        self._description_cached = description

    def __str__(self):
        return f"{self.__class__.__name__}: {self.id}"

    def __init_subclass__(cls, type_id=None):
        if type_id in Dataset.REGISTRY:
            logger.warning(
                f"Type id '{type_id}' already provided by {Dataset.REGISTRY[type_id].__name__}, but {cls.__name__} suggests itself as well"
            )
        if type_id is not None:
            logger.debug(f"Registering specialist {cls.__name__} for type id {type_id}")
            Dataset.REGISTRY[type_id] = cls
        cls.type_id = type_id

    @property
    def is_volume(self):
        """Return True if this dataset represents a brain volume source.

        This property is overwritten by siibra.volumes.VolumeSrc
        """
        return False

    @property
    def is_surface(self):
        """Return True if this dataset represents a brain volume surface. 
        
        This property is overwritten by siibra.volumes.VolumeSrc
        """
        return False

    @property
    def publications(self):
        """
        List of publications for this dataset.
        Empty list here, but implemented in some derived classes.
        """
        return []

    @property
    def urls(self):
        """
        List of URLs related to this dataset.
        Empty list here, but implemented in some derived classes.
        """
        return []

    @property
    def description(self):
        """
        Textual description of Dataset.
        Empty string here, but implemented in some derived classes.
        """
        return self._description_cached

    @classmethod
    def extract_type_id(cls, spec):
        for key in ["@type", "kgSchema"]:
            if key in spec:
                return spec[key]
        raise RuntimeError(f"No type defined in dataset specification: {spec}")


class OriginDescription(Dataset, type_id="fzj/tmp/simpleOriginInfo/v0.0.1"):
    def __init__(self, name, description, urls):
        Dataset.__init__(self, None, description=description)
        # we model the following as property functions,
        # so derived classes may replace them with a lazy loading mechanism.
        self.name = name
        self._urls = urls

    @property
    def urls(self):
        return self._urls

    @classmethod
    def _from_json(cls, spec):
        type_id = cls.extract_type_id(spec)
        assert type_id == cls.type_id
        return cls(
            name=spec["name"],
            description=spec.get("description"),
            urls=spec.get("url", []),
        )


class EbrainsDataset(Dataset, type_id="minds/core/dataset/v1.0.0"):
    def __init__(self, id, name, embargo_status=None):
        Dataset.__init__(self, id, description=None)
        self.embargo_status = embargo_status
        self._name_cached = name
        self._detail = None
        if id is None:
            raise TypeError("Dataset id is required")

        match = re.search(r"([a-f0-9-]+)$", id)
        if not match:
            raise ValueError(
                f"{self.__class__.__name__} initialized with invalid id: {self.id}"
            )
        self._detail_loader = EbrainsKgQuery(
            query_id="interactiveViewerKgQuery-v1_0",
            instance_id=match.group(1),
            params={"vocab": "https://schema.hbp.eu/myQuery/"},
        )

    @property
    def detail(self):
        return self._detail_loader.data

    @property
    def name(self):
        if self._name_cached is None:
            self._name_cached = self.detail.get("name")
        return self._name_cached

    @property
    def publications(self):
        return self.detail.get('publications')

    @property
    def urls(self):
        return [
            {
                "doi": f,
            }
            for f in self.detail.get("kgReference", [])
        ]

    @property
    def description(self):
        return self.detail.get("description")

    @property
    def contributors(self):
        return self.detail.get("contributors")

    @property
    def custodians(self):
        return self.detail.get("custodians")

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, o: object) -> bool:
        if type(o) is not EbrainsDataset and not issubclass(type(o), EbrainsDataset):
            return False
        return self.id == o.id

    @classmethod
    def _from_json(cls, spec):
        type_id = cls.extract_type_id(spec)
        assert type_id == cls.type_id
        return cls(
            id=spec.get("kgId"),
            name=spec.get("name"),
            embargo_status=spec.get("embargo_status", None),
        )



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .requests import DECODERS, HttpRequest, EbrainsRequest, SiibraHttpRequestError
from .. import logger
from abc import ABC, abstractmethod
from urllib.parse import quote
from tqdm import tqdm


class RepositoryConnector(ABC):
    """
    Base class for repository connectors.
    """

    def __init__(self, base_url):
        self.base_url = base_url

    @abstractmethod
    def search_files(folder: str, suffix: str, recursive: bool = False):
        pass

    @abstractmethod
    def _build_url(self, folder: str, filename: str):
        pass

    def _decode_response(self, response, filename):
        # see if we find a default encoder
        suitable_decoders = [
            dec for sfx, dec in DECODERS.items() if filename.endswith(sfx)
        ]
        if len(suitable_decoders) > 0:
            assert len(suitable_decoders) == 1
            return suitable_decoders[0](response)
        else:
            return response

    def get(self, filename, folder="", decode_func=None):
        """Get a file right away."""
        return self.get_loader(filename, folder, decode_func).data

    def get_loader(self, filename, folder="", decode_func=None):
        """Get a lazy loader for a file, for executing the query
        only once loader.data is accessed."""
        url = self._build_url(folder, filename)
        if url is None:
            raise RuntimeError(f"Cannot build url for ({folder}, {filename})")
        if decode_func is None:
            return HttpRequest(url, lambda b: self._decode_response(b, filename))
        else:
            return HttpRequest(url, decode_func)

    def get_loaders(
        self, folder="", suffix=None, progress=None, recursive=False, decode_func=None
    ):
        """
        Returns an iterator with lazy loaders for the files in a given folder.
        In each iteration, a tuple (filename,file content) is returned.
        """
        fnames = self.search_files(folder, suffix, recursive)
        result = [
            (fname, self.get_loader(fname, decode_func=decode_func)) for fname in fnames
        ]
        all_cached = all(_[1].cached for _ in result)
        if progress is None or all_cached:
            return result
        else:
            return tqdm(
                result, total=len(fnames), desc=progress, disable=logger.level > 20
            )


class GitlabConnector(RepositoryConnector):
    def __init__(self, server: str, project: int, reftag: str, skip_branchtest=False):
        # TODO: the query builder needs to check wether the reftag is a branch, and then not cache.
        assert server.startswith("http")
        RepositoryConnector.__init__(
            self, base_url=f"{server}/api/v4/projects/{project}/repository"
        )
        self.reftag = reftag
        self._per_page = 100
        self._branchloader = HttpRequest(
            f"{self.base_url}/branches", DECODERS[".json"], refresh=True
        )
        self._tag_checked = True if skip_branchtest else False
        self._want_commit_cached = None

    def __str__(self):
        return f"{self.__class__.__name__} {self.base_url} {self.reftag}"

    @property
    def want_commit(self):
        if not self._tag_checked:
            try:
                matched_branches = list(
                    filter(lambda b: b["name"] == self.reftag, self.branches)
                )
                if len(matched_branches) > 0:
                    self._want_commit_cached = matched_branches[0]["commit"]
                    logger.debug(
                        f"{self.reftag} is a branch of {self.base_url}! Want last commit "
                        f"{self._want_commit_cached['short_id']} from "
                        f"{self._want_commit_cached['created_at']}"
                    )
                self._tag_checked = True
            except Exception as e:
                print(str(e))
                print("Could not connect to gitlab server!")
        return self._want_commit_cached

    @property
    def branches(self):
        return self._branchloader.data

    def _build_url(self, folder="", filename=None, recursive=False, page=1):
        ref = self.reftag if self.want_commit is None else self.want_commit["short_id"]
        if filename is None:
            pathstr = "" if len(folder) == 0 else f"&path={quote(folder,safe='')}"
            return f"{self.base_url}/tree?ref={ref}{pathstr}&per_page={self._per_page}&page={page}&recursive={recursive}"
        else:
            pathstr = filename if folder == "" else f"{folder}/{filename}"
            filepath = quote(pathstr, safe="")
            return f"{self.base_url}/files/{filepath}/raw?ref={ref}"

    def search_files(self, folder="", suffix=None, recursive=False):
        page = 1
        results = []
        while True:
            loader = HttpRequest(
                self._build_url(folder, recursive=recursive, page=page),
                DECODERS[".json"],
            )
            results.extend(loader.data)
            if len(loader.data) < self._per_page:
                # no more pages
                break
            page += 1
        end = "" if suffix is None else suffix
        return [
            e["path"]
            for e in results
            if e["type"] == "blob" and e["name"].endswith(end)
        ]


class OwncloudConnector(RepositoryConnector):
    def __init__(self, server: str, share: int):
        RepositoryConnector.__init__(self, base_url=f"{server}/s/{share}")

    def search_files(self, folder="", suffix=None, recursive=False):
        raise NotImplementedError(
            f"File search in folders not implemented for {self.__class__.__name__}."
        )

    def _build_url(self, folder, filename):
        fpath = "" if folder == "" else f"path={quote(folder,safe='')}&"
        fpath += f"files={quote(filename)}"
        url = f"{self.base_url}/download?{fpath}"
        return url


class EbrainsHdgConnector(RepositoryConnector):
    """Download sensitive files from EBRAINS using
    the Human Data Gateway (HDG) via the data proxy API.
    """

    base_url = "https://data-proxy.ebrains.eu/api/datasets"
    maxentries = 1000

    def __init__(self, dataset_id):
        """Construct a dataset query for the Human Data Gateway.

        Parameters
        ----------
        dataset_id : str
            EBRAINS dataset id for a dataset that is exposed
            via the human data gateway.
        """

        self._files = []
        self.dataset_id = dataset_id

        marker = None
        while True:

            # The endpoint implements basic pagination, using the filenames as markers.

            if marker is None:
                url = f"{self.base_url}/{dataset_id}?limit={self.maxentries}"
            else:
                url = f"{self.base_url}/{dataset_id}?limit={self.maxentries}&marker={marker}"

            try:
                result = EbrainsRequest(url, DECODERS[".json"]).get()
            except SiibraHttpRequestError as e:
                if e.response.status_code in [401, 422]:
                    # Request access to the dataset (401: expired, 422: not yet requested)
                    EbrainsRequest(f"{self.base_url}/{dataset_id}", post=True).get()
                    input(
                        "You should have received an email with a confirmation link - "
                        "please find that email and click on the link, then press enter "
                        "to continue"
                    )
                    continue
                else:
                    raise RuntimeError(
                        f"Could not request private file links for dataset {dataset_id}. "
                        f"Status code was: {e.response.status_code}. "
                        f"Message was: {e.response.text}. "
                    )

            newfiles = result["objects"]
            self._files.extend(newfiles)
            logger.debug(f"{len(newfiles)} of {self.maxentries} objects returned.")

            if len(newfiles) == self.maxentries:
                # there might be more files
                marker = newfiles[-1]["name"]
            else:
                logger.info(
                    f"{len(self._files)} objects found for dataset {dataset_id} returned."
                )
                self.container = result["container"]
                self.prefix = result["prefix"]
                break

    def search_files(self, folder="", suffix=None, recursive=False):
        result = []
        for f in self._files:
            if f["name"].startswith(folder):
                if suffix is None:
                    result.append(f["name"])
                else:
                    if f["name"].endswith(suffix):
                        result.append(f["name"])
        return result

    def _build_url(self, folder, filename):
        if len(folder) > 0:
            fpath = quote(f"{folder}/{filename}", safe="")
        else:
            fpath = quote(f"{filename}", safe="")
        url = f"{self.base_url}/{self.dataset_id}/{fpath}?redirect=true"
        return url

    def get_loader(self, filename, folder="", decode_func=None):
        """Get a lazy loader for a file, for executing the query
        only once loader.data is accessed."""
        return EbrainsRequest(self._build_url(folder, filename), decode_func)


class EbrainsPublicDatasetConnector(RepositoryConnector):
    """Access files from public EBRAINS datasets via the Knowledge Graph v3 API."""

    QUERY_ID = "bebbe365-a0d6-41ea-9ff8-2554c15f70b7"
    base_url = "https://core.kg.ebrains.eu/v3-beta/queries/"
    maxentries = 1000

    def __init__(self, dataset_id, in_progress=False):
        """Construct a dataset query with the dataset id.

        Parameters
        ----------
        dataset_id : str
            EBRAINS dataset id of a public dataset in KG v3.
        in_progress: bool (default:False)
            If true, will request datasets that are still under curation.
            Will only work when autenticated with an appropriately privileged
            user account.
        """
        self.dataset_id = dataset_id
        stage = "IN_PROGRESS" if in_progress else "RELEASED"
        url = f"{self.base_url}/{self.QUERY_ID}/instances?stage={stage}&dataset_id={dataset_id}"
        result = EbrainsRequest(url, DECODERS[".json"]).get()
        self.versions = {}
        self.use_version = None
        assert len(result["data"]) < 2
        if len(result["data"]) == 1:
            data = result["data"][0]
            self.versions = {v["versionIdentifier"]: v for v in data["versions"]}
            self.use_version = sorted(list(self.versions.keys()))[-1]
            if len(self.versions) > 1:
                logger.info(
                    f"Found {len(self.versions)} versions for dataset '{data['name']}' "
                    f"({', '.join(self.versions.keys())}). "
                    f"Will use {self.use_version} per default."
                )

    @property
    def doi(self):
        if self.use_version in self.versions:
            return self.versions[self.use_version]["doi"]
        else:
            return None

    @property
    def _files(self):
        if self.use_version in self.versions:
            return {
                f["name"]: f["url"] for f in self.versions[self.use_version]["files"]
            }
        else:
            return {}

    def search_files(self, folder="", suffix=None, recursive=False):
        result = []
        for fname in self._files:
            if fname.startswith(folder):
                if suffix is None:
                    result.append(fname)
                else:
                    if fname.endswith(suffix):
                        result.append(fname)
        return result

    def _build_url(self, folder, filename):
        fpath = f"{folder}/{filename}" if len(folder) > 0 else f"{filename}"
        if fpath not in self._files:
            raise RuntimeError(
                f"The file {fpath} requested from EBRAINS dataset {self.dataset_id} cannot be."
            )
        return self._files[fpath]

    def get_loader(self, filename, folder="", decode_func=None):
        """Get a lazy loader for a file, for executing the query
        only once loader.data is accessed."""
        return HttpRequest(self._build_url(folder, filename), decode_func)



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .concept import AtlasConcept
from .atlas import Atlas
from .space import Space, Point, PointSet, BoundingBox
from .region import Region
from .parcellation import Parcellation
from .datasets import Dataset, OriginDescription, EbrainsDataset

# initialize the core concepts and their bootstrapped registries
spaces = Space.REGISTRY
parcellations = Parcellation.REGISTRY
atlases = Atlas.REGISTRY



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ctypes import ArgumentError
from ..commons import logger, Registry, QUIET
from ..core.concept import AtlasConcept
from ..core.atlas import Atlas
from ..core.space import Space, Location, Point, PointSet, BoundingBox
from ..core.region import Region
from ..core.parcellation import Parcellation

from typing import Tuple
from abc import ABC, abstractmethod

try:
    from importlib import resources
except ImportError:
    import importlib_resources as resources
import json
from enum import Enum


class MatchQualification(Enum):
    # Anatomical location of feature matches atlas concept exactly
    EXACT = 0
    # Anatomical location of feature matches atlas concept only approximately
    APPROXIMATE = 1
    # Anatomical location of the feature represents a part of the matched atlas concept
    CONTAINED = 2
    # Anatomical location of the feature inludes the matched atlas concept as a part
    CONTAINS = 3

    @classmethod
    def from_string(cls, name):
        for v in range(4):
            if cls(v).name == name.upper():
                return cls(v)
        return None

    def __str__(self):
        return self.name.lower()


class Match:
    """Description of how a feature has been mapped to an atlas concept."""

    def __init__(
        self,
        concept: AtlasConcept,
        qualification: MatchQualification = MatchQualification.EXACT,
        comment: str = None,
    ):
        """Construct a feature match description.

        Parameters
        ----------
        concept : AtlasConcept
            The atlas concept to which the feature had been mapped.
        qualification: MatchQualifiction
            Qualification of the match, see siibra.feature.MatchQualification
        comment: str
            Optional human-readable explanation comment about the way the match
            was computed
        """
        self.concept = concept
        self.qualification = qualification
        self._comments = []
        if comment is not None:
            self.add_comment(comment)

    @property
    def region(self):
        if isinstance(self.concept, Region):
            return self.concept
        else:
            return None

    @property
    def parcellation(self):
        if isinstance(self.concept, Region):
            return self.concept.parcellation
        elif isinstance(self.concept, Parcellation):
            return self.concept
        else:
            return None

    @property
    def location(self):
        if isinstance(self.concept, Location):
            return self.concept
        else:
            return None

    @property
    def comments(self):
        return ". ".join(self._comments)

    def __str__(self):
        return (
            f"Matched to {self.concept.__class__.__name__} "
            f"'{self.concept.name}' as '{self.qualification}'. "
            f"{self.comments}"
        )

    def add_comment(self, comment: str):
        self._comments.append(comment)


class Feature(ABC):
    """
    Abstract base class for all data features.
    """

    REGISTRY = Registry()

    def __init__(self):
        self._match = None

    def __init_subclass__(cls):
        """
        Registers all subclasses of Feature.
        """
        logger.debug(
            f"Registering feature type {cls.__name__} with modality {cls.modality()}"
        )
        cls.REGISTRY.add(cls.modality(), cls)

    @property
    def matched(self):
        return self._match is not None

    @property
    def match_qualification(self):
        """If this feature was matched agains an atlas concept,
        return the qualification rating of the match.

        Return
        ------
        siibra.features.feature.MatchQualification, if feature was matche
        else None
        """
        return self._match.qualification if self.matched else None

    @property
    def match_description(self):
        return str(self._match) if self.matched else None

    @property
    def matched_region(self):
        return self._match.region if self.matched else None

    @property
    def matched_parcellation(self):
        return self._match.parcellation if self.matched else None

    @property
    def matched_location(self):
        return self._match.location if self.matched else None

    @abstractmethod
    def match(self, concept):
        """
        Matches this feature to the given atlas concept (or a subconcept of it),
        and remembers the matching result.

        Parameters:
        -----------
        concept : AtlasConcept

        Returns:
        -------
        True, if match was successful, otherwise False
        """
        raise RuntimeError(
            f"match() needs to be implemented by derived classes of {self.__class__.__name__}"
        )

    def __str__(self):
        return f"{self.__class__.__name__} feature"

    @classmethod
    def modality(cls):
        """Returns a string representing the modality of a feature."""
        return str(cls).split("'")[1].split(".")[-1]


class SpatialFeature(Feature):
    """
    Base class for coordinate-anchored data features.
    """

    def __init__(self, location: Location):
        """
        Initialize a new spatial feature.

        Parameters
        ----------
        location : Location type
            The location, see siibra.core.location
        """
        assert location is not None
        Feature.__init__(self)
        self.location = location

    @property
    def space(self):
        return self.location.space

    def match(self, concept):
        """
        Matches this feature to the given atlas concept (or a subconcept of it),
        and remembers the matching result.

        TODO this could use parameters for continuous maps, thresholding and resolution

        Parameters:
        -----------
        concept : AtlasConcept

        Returns:
        -------
        True, if match was successful, otherwise False
        """

        self._match = None
        if self.location is None:
            return False

        if isinstance(concept, Parcellation):
            region = concept.regiontree
            logger.info(
                f"{self.__class__} matching against root node {region.name} of {concept.name}"
            )
        elif isinstance(concept, Region):
            region = concept
        else:
            logger.warning(
                f"{self.__class__} cannot match against {concept.__class__} concepts"
            )
            return False

        for tspace in [self.space] + region.supported_spaces:
            if tspace.is_surface:
                continue
            if region.mapped_in_space(tspace):
                if tspace == self.space:
                    return self._test_mask(self.location, region, tspace)
                else:
                    logger.warning(
                        f"{self.__class__.__name__} cannot be tested for {region.name} "
                        f"in {self.space}, testing in {tspace} instead."
                    )
                    return self._test_mask(self.location.warp(tspace), region, tspace)
        else:
            logger.warning(f"Cannot test overlap of {self.location} with {region}")

        return self.matched

    def _test_mask(self, location: Location, region: Region, space: Space):
        mask = region.build_mask(space=space)
        intersection = location.intersection(mask)
        if intersection is None:
            return self.matched
        elif isinstance(location, Point):
            self._match = Match(
                region,
                MatchQualification.EXACT,
                f"Location {location} is inside mask of {region.name}",
            )
        elif isinstance(location, PointSet):
            npts = 1 if isinstance(intersection, Point) else len(intersection)
            if npts == len(location):
                self._match = Match(
                    region,
                    MatchQualification.EXACT,
                    f"All points of {location} inside mask of {region.name}",
                )
            else:
                self._match = Match(
                    region,
                    MatchQualification.APPROXIMATE,
                    f"{npts} of {len(location)} points "
                    f"were inside mask of {region.name}",
                )
        elif isinstance(location, BoundingBox):
            # the intersection of a bounding box with a mask will be a pointset of the
            # mask pixels in the bounding box.
            if location.volume <= intersection.boundingbox.volume:
                self._match = Match(
                    region,
                    MatchQualification.EXACT,
                    f"{str(location)} is fully located inside mask "
                    f"of region {region.name}. ",
                )
            else:
                self._match = Match(
                    region,
                    MatchQualification.APPROXIMATE,
                    f"{str(location)} overlaps with mask " f"of region {region.name}.",
                )
        else:
            self._match = Match(
                region,
                MatchQualification.APPROXIMATE,
                f"Location {location} intersected mask of {region.name}",
            )
        if self.location.space != location.space:
            self._match.add_comment(
                f"The {type(location)} has been warped from {self.location.space} "
                f"to {location.space} for performing the test."
            )
        return self.matched

    def __str__(self):
        return f"{self.__class__.__name__} at {str(self.location)}"


class RegionalFeature(Feature):
    """
    Base class for region-anchored data features (semantic anchoring to region
    names instead of coordinates).
    """

    # load region name aliases from data file
    _aliases = {}
    for species_name in ["human"]:
        species_id = Atlas.get_species_data("human")["@id"]
        with resources.open_text(
            "siibra.features", f"region_aliases_{species_name}.json"
        ) as f:
            _aliases[species_id] = {
                d["Region specification"]
                .lower()
                .strip(): {k: v for k, v in d.items() if k != "Region specification"}
                for d in json.load(f)
            }
        logger.debug(
            f"Loaded {len(_aliases[species_id])} region spec aliases for {species_name}"
        )

    def __init__(self, regionspec: Tuple[str, Region], species=[], **kwargs):
        """
        Parameters
        ----------
        regionspec : string or Region
            Specifier for the brain region, will be matched at test time
        """
        if not any(map(lambda c: isinstance(regionspec, c), [Region, str])):
            raise TypeError(
                f"invalid type {type(regionspec)} provided as region specification"
            )
        Feature.__init__(self)
        self.regionspec = regionspec
        if isinstance(species, list):
            self.species = species
        elif isinstance(species, dict):
            self.species = [species]
        else:
            raise ArgumentError(
                f"Type {type(species)} not expected for species, should be list or dict."
            )

    @property
    def species_ids(self):
        return [s.get("@id") for s in self.species]

    def match(self, concept):
        """
        Matches this feature to the given atlas concept (or a subconcept of it),
        and remembers the matching result.

        Parameters:
        -----------
        concept : AtlasConcept

        Returns:
        -------
        True, if match was successful, otherwise False
        """
        self._match = None

        # Verify that a possible species specification of the feature matches the
        # given concept at all.
        try:
            if isinstance(concept, Region):
                atlases = concept.parcellation.atlases
            if isinstance(concept, Parcellation):
                atlases = concept.atlases
            if isinstance(concept, Atlas):
                atlases = {concept}
            if atlases:
                # if self.species_ids is defined, and the concept is explicitly not in
                # return False
                if all(
                    atlas.species.get("@id") not in self.species_ids
                    for atlas in atlases
                ):
                    return self.matched
        # for backwards compatibility. If any attr is not found, pass
        except AttributeError:
            pass

        # Feature's region is specified as a Region object
        # -> we can apply simple tests for any query object.
        if isinstance(self.regionspec, Region):
            return self._match_region(self.regionspec, concept)

        assert isinstance(self.regionspec, str)
        spec = self.regionspec.lower().strip()

        # Feature's region is specified by string. Check alias table first.
        for species_id in set(self.species_ids) & set(self._aliases.keys()):
            if spec in self._aliases[species_id]:
                repl = self._aliases[species_id][spec]
                with QUIET:
                    if repl["Matched parcellation"] is not None:
                        # a matched parcellation is stored in the alias table, this will be preferred.
                        parc = Parcellation.REGISTRY[repl["Matched parcellation"]]
                        spec = repl["Matched region"]
                        msg_alias = (
                            f"Original region specification was '{self.regionspec}'."
                        )
                    else:
                        # no matched parcellation, then we use the original one given in the table.
                        assert repl["Origin parcellation"] is not None
                        parc = Parcellation.REGISTRY[repl["Origin parcellation"]]
                        msg_alias = (
                            f"Parcellation '{parc}' was used to decode '{spec}'."
                        )
                    for r in [parc.decode_region(s) for s in spec.split(",")]:
                        if self._match_region(r, concept):
                            break
                if self._match is not None:
                    self._match.add_comment(msg_alias)
                    if repl["Qualification"] is not None:
                        self._match.qualification = MatchQualification.from_string(
                            repl["Qualification"]
                        )
                return self.matched

        # Feature's region is specified by string, search query is a Parcellation
        if isinstance(concept, Parcellation):
            logger.debug(
                f"{self.__class__} matching against root node {concept.regiontree.name} of {concept.name}"
            )
            for w in concept.key.split("_"):
                spec = spec.replace(w.lower(), "")
            for region in concept.regiontree.find(spec):
                self._match = Match(
                    region,
                    MatchQualification.CONTAINED,
                    f"Feature was linked to {region.name}, which belongs to parcellation {concept.name}.",
                )
                return True

        # Feature's region is specified by string, search query is a Region
        elif isinstance(concept, Region):
            for w in concept.parcellation.key.split("_"):
                if not w.isnumeric() and len(w) > 2:
                    spec = spec.replace(w.lower(), "")
            for region in concept.find(spec):
                if region == concept:
                    self._match = Match(region, MatchQualification.EXACT)
                else:
                    self._match = Match(
                        concept,
                        MatchQualification.CONTAINED,
                        f"Feature was linked to child region '{region.name}'",
                    )
                return True

        # Feature's region is specified by string, search query is an Atlas
        elif isinstance(concept, Atlas):
            logger.debug(
                "Matching regional features against a complete atlas. "
                "This is not efficient and the query may take a while."
            )
            for w in concept.key.split("_"):
                spec = spec.replace(w.lower(), "")
            for p in concept.parcellations:
                for region in p.regiontree.find(spec):
                    self._match = Match(
                        concept,
                        MatchQualification.CONTAINED,
                        f"Region {region.name} belongs to atlas parcellation {p.name}",
                    )
                    return True
        else:
            logger.warning(
                f"{self.__class__} cannot match against {concept.__class__} concepts"
            )

        return self.matched

    def _match_region(self, region: Region, concept: AtlasConcept):
        """
        Match a decoded region object representing the anatomical location
        of this feature to a given atlas concept.

        This is only a convenience function used by match(), and not
        meant for direct evaluation.
        """
        self._match = None
        if isinstance(concept, Parcellation):
            if region in concept:
                self._match = Match(
                    concept,
                    MatchQualification.CONTAINED,
                    f"Feature belongs to {region.name}, which is part of "
                    f"parcellation {concept.name}.",
                )
        elif isinstance(concept, Region):
            if region == concept:
                self._match = Match(concept, MatchQualification.EXACT)
            elif region.has_parent(concept):
                self._match = Match(
                    concept,
                    MatchQualification.CONTAINED,
                    f"Feature was linked to child region '{region.name}'",
                )
            elif concept.has_parent(region):
                self._match = Match(
                    concept,
                    MatchQualification.CONTAINS,
                    f"Feature was linked to parent region '{region.name}'",
                )
        elif isinstance(concept, Atlas):
            if any(region in p for p in concept.parcellations):
                self._match = Match(
                    concept,
                    MatchQualification.CONTAINED,
                    f"Feature belongs to {region.name}, which is part of "
                    f"a parcellation supported by atlas {concept.name}.",
                )
        return self.matched

    def __str__(self):
        return f"{self.__class__.__name__} for {self.regionspec}"


class ParcellationFeature(Feature):
    """
    Base class for data features which apply to the atlas as a whole
    instead of a particular location or region. A typical example is a
    connectivity matrix, which applies to all regions in the atlas.
    """

    def __init__(self, parcellationspec):
        """
        Parameters
        ----------
        parcellationspec : str or Parcellation object
            Identifies the underlying parcellation
        """
        Feature.__init__(self)
        self.spec = parcellationspec
        self.parcellations = Parcellation.REGISTRY.find(parcellationspec)

    def match(self, concept):
        """
        Matches this feature to the given atlas concept (or a subconcept of it),
        and remembers the matching result.

        Parameters:
        -----------
        concept : AtlasConcept

        Returns:
        -------
        True, if match was successful, otherwise False
        """
        self._match = None
        if isinstance(concept, Parcellation):
            if concept in self.parcellations:
                self._match = concept
        elif isinstance(concept, Region):
            if concept.parcellation in self.parcellations:
                self._match = concept
        elif isinstance(concept, Atlas):
            logger.debug(
                "Matching a parcellation feature against a complete atlas. "
                "This will return features matching any supported parcellation, "
                "including different parcellation versions."
            )
            for p in concept.parcellations:
                if p in self.parcellations:
                    self._match = p
                    return True
        else:
            logger.warning(
                f"{self.__class__} cannot match against {concept.__class__} concepts"
            )

        return self.matched

    def __str__(self):
        return f"{self.__class__.__name__} for {self.spec}"



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .feature import RegionalFeature
from .query import FeatureQuery

from ..commons import logger
from ..retrieval.requests import EbrainsKgQuery, HttpRequest
from ..core.datasets import EbrainsDataset

import PIL.Image as Image
import numpy as np
from io import BytesIO
from collections import namedtuple
import re
import importlib


RECEPTOR_SYMBOLS = {
    "5-HT1A": {
        "receptor": {
            "latex": "$5-HT_{1A}$",
            "markdown": "5-HT<sub>1A</sub>",
            "name": "5-hydroxytryptamine receptor type 1A",
        },
        "neurotransmitter": {
            "label": "5-HT",
            "latex": "$5-HT$",
            "markdown": "5-HT",
            "name": "serotonin",
        },
    },
    "5-HT2": {
        "receptor": {
            "latex": "$5-HT_2$",
            "markdown": "5-HT<sub>2</sub>",
            "name": "5-hydroxytryptamine receptor type 2",
        },
        "neurotransmitter": {
            "label": "5-HT",
            "latex": "$5-HT$",
            "markdown": "5-HT",
            "name": "serotonin",
        },
    },
    "AMPA": {
        "receptor": {
            "latex": "$AMPA$",
            "markdown": "AMPA",
            "name": "alpha-amino-3hydroxy-5-methyl-4isoxazolepropionic acid receptor",
        },
        "neurotransmitter": {
            "label": "Glu",
            "latex": "$Glu$",
            "markdown": "Glu",
            "name": "glutamate",
        },
    },
    "BZ": {
        "receptor": {
            "latex": "$GABA_A/BZ$",
            "markdown": "GABA<sub>A</sub>/BZ",
            "name": "gamma-aminobutyric acid receptor type A / benzodiazepine associated binding site",
        },
        "neurotransmitter": {
            "label": "GABA",
            "latex": "$GABA$",
            "markdown": "GABA",
            "name": "gamma-aminobutyric acid",
        },
    },
    "D1": {
        "receptor": {
            "latex": "$D_1$",
            "markdown": "D<sub>1</sub>",
            "name": "dopamine receptor type 1",
        },
        "neurotransmitter": {
            "label": "DA",
            "latex": "$DA$",
            "markdown": "DA",
            "name": "dopamine",
        },
    },
    "GABAA": {
        "receptor": {
            "latex": "$GABA_A$",
            "markdown": "GABA<sub>A</sub>",
            "name": "gamma-aminobutyric acid receptor type A",
        },
        "neurotransmitter": {
            "label": "GABA",
            "latex": "$GABA$",
            "markdown": "GABA",
            "name": "gamma-aminobutyric acid",
        },
    },
    "GABAB": {
        "receptor": {
            "latex": "$GABA_B$",
            "markdown": "GABA<sub>B</sub>",
            "name": "gamma-aminobutyric acid receptor type B",
        },
        "neurotransmitter": {
            "label": "GABA",
            "latex": "$GABA$",
            "markdown": "GABA",
            "name": "gamma-aminobutyric acid",
        },
    },
    "M1": {
        "receptor": {
            "latex": "$M_1$",
            "markdown": "M<sub>1</sub>",
            "name": "muscarinic acetylcholine receptor type 1",
        },
        "neurotransmitter": {
            "label": "ACh",
            "latex": "$ACh$",
            "markdown": "ACh",
            "name": "acetylcholine",
        },
    },
    "M2": {
        "receptor": {
            "latex": "$M_2$",
            "markdown": "M<sub>2</sub>",
            "name": "muscarinic acetylcholine receptor type 2",
        },
        "neurotransmitter": {
            "label": "ACh",
            "latex": "$ACh$",
            "markdown": "ACh",
            "name": "acetylcholine",
        },
    },
    "M3": {
        "receptor": {
            "latex": "$M_3$",
            "markdown": "M<sub>3</sub>",
            "name": "muscarinic acetylcholine receptor type 3",
        },
        "neurotransmitter": {
            "label": "ACh",
            "latex": "$ACh$",
            "markdown": "ACh",
            "name": "acetylcholine",
        },
    },
    "NMDA": {
        "receptor": {
            "latex": "$NMDA$",
            "markdown": "NMDA",
            "name": "N-methyl-D-aspartate receptor",
        },
        "neurotransmitter": {
            "label": "Glu",
            "latex": "$Glu$",
            "markdown": "Glu",
            "name": "glutamate",
        },
    },
    "alpha1": {
        "receptor": {
            "latex": "$\\alpha_1$",
            "markdown": "&#945<sub>1</sub>",
            "name": "alpha-1 adrenergic receptor",
        },
        "neurotransmitter": {
            "label": "NE",
            "latex": "$NE$",
            "markdown": "NE",
            "name": "norepinephrine",
        },
    },
    "alpha2": {
        "receptor": {
            "latex": "$\\alpha_2$",
            "markdown": "&#945<sub>2</sub>",
            "name": "alpha-2 adrenergic receptor",
        },
        "neurotransmitter": {
            "label": "NE",
            "latex": "$NE$",
            "markdown": "NE",
            "name": "norepinephrine",
        },
    },
    "alpha4beta2": {
        "receptor": {
            "latex": "$\\alpha_4\\beta_2$",
            "markdown": "&#945<sub>4</sub>&#946<sub>2</sub>",
            "name": "alpha-4 beta-2 nicotinic receptor",
        },
        "neurotransmitter": {
            "label": "ACh",
            "latex": "$ACh$",
            "markdown": "ACh",
            "name": "acetylcholine",
        },
    },
    "kainate": {
        "receptor": {
            "latex": "$kainate$",
            "markdown": "kainate",
            "name": "kainate receptors",
        },
        "neurotransmitter": {
            "label": "Glu",
            "latex": "$Glu$",
            "markdown": "Glu",
            "name": "glutamate",
        },
    },
    "mGluR2_3": {
        "receptor": {
            "latex": "$mGluR2/3$",
            "markdown": "mGluR2/3",
            "name": "metabotropic glutamate receptor type 2 and 3",
        },
        "neurotransmitter": {
            "label": "Glu",
            "latex": "$Glu$",
            "markdown": "Glu",
            "name": "glutamate",
        },
    },
}


def unify_stringlist(L: list):
    """Adds asterisks to strings that appear multiple times, so the resulting
    list has only unique strings but still the same length, order, and meaning.
    For example:
        unify_stringlist(['a','a','b','a','c']) -> ['a','a*','b','a**','c']
    """
    assert all([isinstance(_, str) for _ in L])
    return [L[i] + "*" * L[:i].count(L[i]) for i in range(len(L))]


def decode_tsv(bytearray):
    bytestream = BytesIO(bytearray)
    header = bytestream.readline()
    lines = [_.strip() for _ in bytestream.readlines() if len(_.strip()) > 0]
    sep = b"{" if b"{" in lines[0] else b"\t"
    keys = unify_stringlist(
        [
            n.decode("utf8").replace('"', "").replace("'", "").strip()
            for n in header.split(sep)
        ]
    )
    if any(k.endswith("*") for k in keys):
        logger.warning("Redundant headers in receptor file")
    assert len(keys) == len(set(keys))
    return {
        _.split(sep)[0].decode("utf8"): dict(
            zip(
                keys,
                [re.sub(r"\\+", r"\\", v.decode("utf8").strip()) for v in _.split(sep)],
            )
        )
        for _ in lines
    }


class DensityProfile:
    def __init__(self, data):
        units = {list(v.values())[3] for v in data.values()}
        assert len(units) == 1
        self.unit = next(iter(units))
        self.densities = {
            int(k): float(list(v.values())[2]) for k, v in data.items() if k.isnumeric()
        }

    def __iter__(self):
        return self.densities.values()


Density = namedtuple("Density", "name, mean, std, unit")


class DensityFingerprint:

    unit = None
    labels = []
    meanvals = []
    stdvals = []
    n = 0

    def __init__(self, datadict):
        """
        Create a DensityFingerprint from a data dictionary coming from a
        receptor fingerprint tsv file.
        """
        units = {list(v.values())[3] for v in datadict.values()}
        assert len(units) == 1
        self.unit = next(iter(units))
        self.labels = list(datadict.keys())
        try:
            mean = [datadict[_]["density (mean)"] for _ in self.labels]
            std = [datadict[_]["density (sd)"] for _ in self.labels]
        except KeyError as e:
            print(str(e))
            logger.error("Could not parse fingerprint from this dictionary")
        self.meanvals = [float(m) if m.isnumeric() else 0 for m in mean]
        self.stdvals = [float(s) if s.isnumeric() else 0 for s in std]

    def __getitem__(self, index):
        if isinstance(index, int):
            if index >= len(self.labels):
                return None
            index_ = index
        elif isinstance(index, str):
            if index not in self.labels:
                return None
            index_ = self.labels.index(index)
        return Density(
            name=self.labels[index_],
            mean=self.meanvals[index_],
            std=self.stdvals[index_],
            unit=self.unit,
        )

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self):
        if self.n < len(self.labels):
            self.n += 1
            return self[self.n - 1]
        else:
            raise StopIteration

    def __str__(self):
        return "\n".join(
            "{d.name:15.15s} {d.mean:8.1f} {d.unit} (+/-{d.std:5.1f})".format(d=d)
            for d in iter(self)
        )


class ReceptorDistribution(RegionalFeature, EbrainsDataset):
    """
    Reprecent a receptor distribution dataset with fingerprint, profiles and
    autoradiograph samples. This implements a lazy loading scheme.
    TODO lazy loading could be more elegant.
    """

    def __init__(self, region, kg_result, **kwargs):

        RegionalFeature.__init__(self, region, **kwargs)
        EbrainsDataset.__init__(self, kg_result["identifier"], kg_result["name"])

        self.info = kg_result["description"]
        self.url = "https://search.kg.ebrains.eu/instances/Dataset/{}".format(self.id)
        self.modality = kg_result["modality"]

        urls = kg_result["files"]
        
        self.files=urls

        def urls_matching(regex):
            return filter(lambda u: re.match(regex, u), urls)

        # add fingerprint if a url is found
        self._fingerprint_loader = None
        for url in urls_matching(".*_fp[._]"):
            if self._fingerprint_loader is not None:
                logger.warning(f"More than one fingerprint found for {self}")
            self._fingerprint_loader = HttpRequest(
                url, lambda u: DensityFingerprint(decode_tsv(u))
            )

        # add any cortical profiles
        self._profile_loaders = {}
        for url in urls_matching(r".*_pr[._].*\.tsv"):
            rtype, basename = url.split("/")[-2:]
            if rtype not in basename:
                continue
            if rtype in self._profile_loaders:
                logger.warning(f"More than one profile for '{rtype}' in {self.url}")
            self._profile_loaders[rtype] = HttpRequest(
                url, lambda u: DensityProfile(decode_tsv(u))
            )

        # add autoradiograph
        self._autoradiograph_loaders = {}

        def img_from_bytes(b):
            return np.array(Image.open(BytesIO(b)))

        for url in urls_matching(".*_ar[._]"):
            rtype, basename = url.split("/")[-2:]
            if rtype not in basename:
                continue
            if rtype in self._autoradiograph_loaders:
                logger.warning(
                    f"More than one autoradiograph for '{rtype}' in {self.url}"
                )
            self._autoradiograph_loaders[rtype] = HttpRequest(url, img_from_bytes)

    @property
    def fingerprint(self):
        """The receptor fingerprint, with mean and standard
        deviations of receptor densities for different
        receptor types, measured across multiple samples."""
        if self._fingerprint_loader is None:
            return None
        else:
            return self._fingerprint_loader.data

    @property
    def profiles(self):
        """Dictionary of cortical receptor distribution
        profiles available for this feature, keyed by
        receptor type."""
        return {rtype: l.data for rtype, l in self._profile_loaders.items()}

    @property
    def autoradiographs(self):
        """Dictionary of sample autoradiographs available for this feature,
        keyed by receptor type."""
        return {rtype: l.data for rtype, l in self._autoradiograph_loaders.items()}

    def plot(self, title=None):
        """
        Produce a fingerprint and cortical profile plot of this receptor distribution feature.
        """

        if importlib.util.find_spec("matplotlib") is None:
            logger.warning("matplotlib not available. Plotting disabled.")
            return None

        from matplotlib import pyplot
        from collections import deque

        # plot profiles and fingerprint
        fig = pyplot.figure(figsize=(12, 4))
        pyplot.subplot(121)
        for _, profile in self.profiles.items():
            pyplot.plot(
                list(profile.densities.keys()),
                np.fromiter(profile.densities.values(), dtype="d"),
            )
        pyplot.xlabel("Cortical depth (%)")
        pyplot.ylabel("Receptor density")
        pyplot.grid(True)
        pyplot.suptitle(str(self) if title is None else title)
        pyplot.legend(
            labels=[_ for _ in self.profiles],
            loc="upper center",
            bbox_to_anchor=(0.5, 1.05),
            ncol=4,
            prop={"size": 8},
        )

        ax = pyplot.subplot(122, projection="polar")
        angles = deque(
            np.linspace(0, 2 * np.pi, len(self.fingerprint.labels) + 1)[:-1][::-1]
        )
        angles.rotate(5)
        angles = list(angles)
        means = [d.mean for d in self.fingerprint]
        stds = [d.mean + d.std for d in self.fingerprint]
        pyplot.plot(angles + [angles[0]], means + [means[0]], "k-", lw=3)
        pyplot.plot(angles + [angles[0]], stds + [stds[0]], "k", lw=1)
        ax.set_xticks(angles)
        ax.set_xticklabels([_ for _ in self.fingerprint.labels])
        ax.tick_params(pad=9, labelsize=10)
        ax.tick_params(axis="y", labelsize=8)
        return fig


class ReceptorQuery(FeatureQuery):

    _FEATURETYPE = ReceptorDistribution

    def __init__(self,**kwargs):
        FeatureQuery.__init__(self)
        kg_req = EbrainsKgQuery(
            query_id="siibra_receptor_densities-0_0_2",
            params={'vocab': 'https://schema.hbp.eu/myQuery/' }
        )
        kg_query = kg_req.get()
        
        not_used = 0
        for kg_result in kg_query["results"]:
            region_names = [p_region["name"] for p_region in kg_result["parcellationRegion"]]
            species = kg_result.get('species', [])
            for region_name in region_names:
                f = ReceptorDistribution(region_name, kg_result, species=species)
                if f.fingerprint is None:
                    not_used += 1
                else:
                    self.register(f)
        
        if not_used > 0:
            logger.info(f'{not_used} receptor datasets skipped due to unsupported format.')



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .feature import SpatialFeature
from .query import FeatureQuery

from .. import logger
from ..core.datasets import EbrainsDataset
from ..core.space import Space, Point, PointSet, WholeBrain
from ..retrieval.repositories import GitlabConnector

import re


class IEEG_Dataset(SpatialFeature, EbrainsDataset):
    """
    A Dataset of intracranial EEG measurements retrieved from EBRAINS,
    composed of different sessions on different subjects.
    """
    def __init__(self, dataset_id, name, space):
        SpatialFeature.__init__(self, WholeBrain(space))
        EbrainsDataset.__init__(self, dataset_id, name)
        self.sessions = {}

    def new_session(self, subject_id):
        # NOTE this will call register_session on construction!
        return IEEG_Session(self, subject_id)

    def register_session(self, s):
        if s.subject_id in self.sessions:
            logger.warning(f"Session {str(s)} alread registered!")
        self.sessions[s.subject_id] = s
        self._update_location()

    def __iter__(self):
        """
        Iterate over sessions
        """
        return iter(self.sessions.values())

    def _update_location(self):
        points = []
        for s in self:
            if s.location is not None:
                points.extend(list(s.location))
        if len(points) > 0:
            self.location = PointSet(points, points[0].space)

    @classmethod
    def _from_json(cls, spec):
        return cls(
            dataset_id=spec["kgId"],
            name=spec["name"],
            space=Space.REGISTRY[spec["space id"]],
        )


class IEEG_Session(SpatialFeature):
    """
    An intracranial EEG recording session on a particular subject,
    storing as set of electrodes and linking to an IEEG_Dataset.
    """

    def __init__(self, dataset: IEEG_Dataset, subject_id):
        SpatialFeature.__init__(self, dataset.location)
        self.sub_id = subject_id
        self.dataset = dataset
        self.electrodes = {}  # key: subject_id

    def new_electrode(self, electrode_id):
        return IEEG_Electrode(
            self, electrode_id
        )  # will call register_electrode on construction!

    def register_electrode(self, e):
        if e.electrode_id in self.electrodes:
            logger.warning(
                "Electrode {e.electrode_id} of {e.subject_id} already registered!"
            )
        self.electrodes[e.electrode_id] = e
        self._update_location()

    def __iter__(self):
        """
        Iterate over electrodes
        """
        return iter(self.electrodes.values())

    def _update_location(self):
        points = []
        for electrode in self:
            if electrode.location is not None:
                points.extend(list(electrode.location))
        if len(points) > 0:
            self.location = PointSet(points, points[0].space)
            self.dataset._update_location()


class IEEG_Electrode(SpatialFeature):
    """
    EEG Electrode with multiple contact points placed in a reference space,
    linking to a particular IEEG recording session.
    """
    def __init__(self, session: IEEG_Session, electrode_id):
        SpatialFeature.__init__(self, session.location)
        self.session = session
        self.electrode_id = electrode_id
        self.contact_points = {}
        session.register_electrode(self)

    def new_contact_point(self, id, coord):
        return IEEG_ContactPoint(
            self, id, coord
        )  # will call register_contact_point on construction!

    def register_contact_point(self, contactpoint):
        if contactpoint.id in self.contact_points:
            raise ValueError(
                f"Contact point with id {contactpoint.id} already registered to {self}"
            )
        self.contact_points[contactpoint.id] = contactpoint
        self._update_location()

    def __iter__(self):
        """
        Iterate over contact points
        """
        return iter(self.contact_points.values())

    def _update_location(self):
        points = [cp.location for cp in self if cp.location is not None]
        if len(points) > 0:
            self.location = PointSet(points, self.session.space)
            self.session._update_location()


class IEEG_ContactPoint(SpatialFeature):
    """
    Basic regional feature for iEEG contact points.
    """

    def __init__(self, electrode, id, coord):
        point = Point(coord, electrode.space)
        SpatialFeature.__init__(self, point)
        self.electrode = electrode
        self.id = id
        electrode.register_contact_point(self)

    def next(self):
        """
        Returns the next contact point of the same electrode, if any.
        """
        ids_available = list(self.electrode.contact_points.keys())
        my_index = ids_available.index(self.id)
        next_index = my_index + 1
        if next_index < len(ids_available):
            next_id = ids_available[next_index]
            return self.electrode.contact_points[next_id]
        else:
            return None

    def prev(self):
        """
        Returns the previous contact point of the same electrode, if any.
        """
        ids_available = list(self.electrode.contact_points.keys())
        my_index = ids_available.index(self.id)
        prev_index = my_index - 1
        if prev_index >= 0:
            prev_id = ids_available[prev_index]
            return self.electrode.contact_points[prev_id]
        else:
            return None


def parse_ptsfile(spec):
    lines = spec.split("\n")
    N = int(lines[2].strip())
    result = {}
    for i in range(N):
        fields = re.split("\t", lines[i + 3].strip())
        electrode_id, contact_point_id = re.split(r"(\d+)", fields[0])[:-1]
        if electrode_id not in result:
            result[electrode_id] = {}
        assert contact_point_id not in result[electrode_id]
        result[electrode_id][contact_point_id] = list(map(float, fields[1:4]))
    return result


class IEEG_SessionQuery(FeatureQuery):
    _FEATURETYPE = IEEG_Session
    _CONNECTOR = GitlabConnector("https://jugit.fz-juelich.de", 3009, "master")

    def __init__(self,**kwargs):

        FeatureQuery.__init__(self)
        dset = IEEG_Dataset._from_json(
            self._CONNECTOR.get_loader("ieeg_contact_points/info.json").data
        )

        for fname, loader in self._CONNECTOR.get_loaders("ieeg_contact_points", ".pts"):

            logger.debug(f"Retrieving from {fname}")

            obj = parse_ptsfile(loader.data.decode())
            subject_id = fname.replace('ieeg_contact_points/', '').split("_")[0]
            session = dset.new_session(subject_id)
            for electrode_id, contact_points in obj.items():
                electrode = session.new_electrode(electrode_id)
                for contact_point_id, coord in contact_points.items():
                    electrode.new_contact_point(contact_point_id, coord)
            self.register(session)


if __name__ == "__main__":
    extractor = IEEG_SessionQuery()



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .feature import ParcellationFeature
from .query import FeatureQuery

from ..commons import logger, QUIET
from ..core.parcellation import Parcellation
from ..retrieval.repositories import GitlabConnector, EbrainsPublicDatasetConnector

import pandas as pd
from typing import List


class ConnectivityMatrix(ParcellationFeature):

    """ Structural connectivity matrix grouped by a parcellation. """
    def __init__(self, parcellation_id: str, matrixloader, srcinfo):
        """Construct a parcellation-averaged connectivty matrix.
        
        Arguments
        ---------
        parcellation_id : str
            Id of corresponding parcellation
        matrixloader : func
            Function which loads the matrix as a pandas dataframe
        """
        ParcellationFeature.__init__(self, parcellation_id)
        self._matrix_loader = matrixloader
        self.src_info = srcinfo

    @property
    def matrix(self):
        # load and return the matrix
        return self._matrix_loader()

    def get_profile(self, regionspec):
        for p in self.parcellations:
            region = p.decode_region(regionspec)
            return self.matrix[region]

    def __str__(self):
        return ParcellationFeature.__str__(self) + " " + str(self.src_info)


class StreamlineCounts(ConnectivityMatrix):
    """ Structural connectivity matrix of streamline counts grouped by a parcellation. """
    def __init__(self, parcellation_id: str, matrixloader, srcinfo):
        super().__init__(parcellation_id, matrixloader, srcinfo)


class StreamlineLengths(ConnectivityMatrix):
    """ Structural connectivity matrix of streamline lengths grouped by a parcellation. """
    def __init__(self, parcellation_id: str, matrixloader, srcinfo):
        super().__init__(parcellation_id, matrixloader, srcinfo)


class FunctionalConnectivity(ConnectivityMatrix):
    """ Functional connectivity matrix, grouped by a parcellation. """
    def __init__(self, parcellation_id: str, matrixloader, paradigm: str, srcinfo):
        super().__init__(parcellation_id, matrixloader, srcinfo)
        self.paradigm = paradigm


class HcpConnectivityFetcher():

    _DATASET_ID = '0f1ccc4a-9a11-4697-b43f-9c9c8ac543e6'
    # TODO add other parcellations
    _PARCELLATION_NAMES = {
        "294-Julich-Brain" : Parcellation.REGISTRY.JULICH_BRAIN_CYTOARCHITECTONIC_MAPS_2_9
    }

    def __init__(self, filename_keyword ):
        FeatureQuery.__init__(self)
        logger.warn(
            f"The connectivity dataset {self._DATASET_ID} is still under curation, "
            "and currently only accessible with a privileged EBRAINS account."
        )
        self._connector = EbrainsPublicDatasetConnector(self._DATASET_ID, in_progress=True)
        self._keyword = filename_keyword

    @property
    def doi(self):
        return self._connector.doi

    @property
    def srcinfo(self):
        return {
            'dataset_id': self._DATASET_ID, 
            'doi': self.doi,
            'cohort': 'HCP'
        }

    def get_matrixloaders(self, parcellation: Parcellation):
        """ Return functions for loading the connectivity matrices 
        matching the given parcellation as pandas DataFrames. """

        loaders = []
        for name, parc in self._PARCELLATION_NAMES.items():

            if parc != parcellation:
                continue

            try:
                zipfile = self._connector.get(f'{name}.zip')
            except RuntimeError as e:
                logger.error(str(e))
                continue
        
            # extract index - regionname mapping
            with zipfile.open(f'{name}/0ImageProcessing/Link.txt') as f:
                lines = [l.decode().strip().split(' ', maxsplit=1) for l in f.readlines()]
                with QUIET:
                    indexmap = {
                        int(l[0]):parc.decode_region(l[1]) 
                        for l in lines 
                        if len(l)==2 and l[0].isnumeric()
                    } 
    
            # create a dict of the csv files by subject id
            csvfiles = [
                f for f in zipfile.namelist() 
                if f.endswith('.csv') and (self._keyword in f)
            ]

            for fn in csvfiles:
                subject_id = fn.split('/')[-2]
                # define the lazy loader function
                loaders.append((
                    fn,
                    subject_id, 
                    lambda z=f'{name}.zip', c=fn, m=indexmap: self._load_matrix(z, c, m)
                ))
                
        return loaders

    def _load_matrix(self, zip_filename, csv_filename, indexmap, separators=", ;"):
        """Extract connectivity matrix from a csv file inside a zip file 
        known to the dataset connector, and return it as a dataframe with 
        region objects as row and column indices.
        """ 
        logger.debug(f"Loading {csv_filename} from {zip_filename}.")
        zipfile = self._connector.get(zip_filename)
        for sep in separators:
            try:
                matrix = pd.read_csv(
                    zipfile.open(csv_filename), sep=sep,
                    header=0, names=list(indexmap.values()), index_col=False
                )
                # set row indices to region names as well
                remapper = {l-1:region for l, region in indexmap.items()}
                matrix.rename(index=remapper, inplace=True)
                return matrix
            except pd.errors.ParserError:
                continue
        logger.error(f"Could not parse connectivity matrix from file {csv_filename} in {zip_filename}.")

    @property
    def parcellations(self):
        return list(self._PARCELLATION_NAMES.values())


class HcpStreamlineCountQuery(HcpConnectivityFetcher, FeatureQuery):

    _FEATURETYPE = StreamlineCounts

    def __init__(self, **kwargs):
        FeatureQuery.__init__(self)
        HcpConnectivityFetcher.__init__(self, filename_keyword="Counts")
        for parc in self.parcellations:
            for filename, subject_id, loader in self.get_matrixloaders(parc):
                srcinfo = self.srcinfo
                srcinfo['subject'] = subject_id
                srcinfo['filename'] = filename
                self.register(self._FEATURETYPE(parc, loader, srcinfo))


class HcpStreamlineLengthQuery(HcpConnectivityFetcher, FeatureQuery):

    _FEATURETYPE = StreamlineLengths

    def __init__(self, **kwargs):
        FeatureQuery.__init__(self)
        HcpConnectivityFetcher.__init__(self, filename_keyword="Lengths")
        for parc in self.parcellations:
            for filename, subject_id, loader in self.get_matrixloaders(parc):
                srcinfo = self.srcinfo
                srcinfo['subject'] = subject_id
                srcinfo['filename'] = filename
                self.register(self._FEATURETYPE(parc, loader, srcinfo))


class HcpRestingStateQuery(HcpConnectivityFetcher, FeatureQuery):

    _FEATURETYPE = FunctionalConnectivity

    def __init__(self, **kwargs):
        FeatureQuery.__init__(self)
        HcpConnectivityFetcher.__init__(self, filename_keyword="EmpCorrFC")
        for parc in self.parcellations:
            for filename, subject_id, loader in self.get_matrixloaders(parc):
                paradigm = "Resting state ({})".format(
                    filename.split('/')[-1].replace('.csv','')
                )
                srcinfo = self.srcinfo
                srcinfo['subject'] = subject_id
                srcinfo['filename'] = filename
                self.register(self._FEATURETYPE(parc, loader, paradigm, srcinfo))


class PrereleasedConnectivityFetcher(GitlabConnector):

    _COHORTS = ['1000brains', 'hcp', 'enki']

    def __init__(self, keywords: List[str] ):
        """Gitlab connector for accessing some pre-released connectivity matrices.
        """
        GitlabConnector.__init__(self, "https://jugit.fz-juelich.de", 3009, "develop")
        self._keywords = keywords

    def get_data(self):
        results = []
        for filename, jsonloader in self.get_loaders("connectivity", ".json"):
            if not any(kw.lower() in filename.lower() for kw in self._keywords):
                continue
            parc_id = jsonloader.data["parcellation id"]
            matrixloader = lambda l=jsonloader: self._matrixloader(l)
            srcinfo = {k:v for k, v in jsonloader.data.items() if k!='data'}
            srcinfo['filename'] = filename
            matched_cohorts = [c for c in self._COHORTS if filename.lower().find(c.lower())>=0]
            if len(matched_cohorts)>0:
                srcinfo['cohort'] = matched_cohorts[0].upper()
                srcinfo['subject'] = 'average'
            results.append((parc_id, matrixloader, srcinfo))
        return results

    def _matrixloader(self, jsonloader):
        """
        Load connectivity matrix from json object
        """
        data = jsonloader.data
        assert 'data' in data
        col_names = data['data']['field names']
        row_names = list(data['data']['profiles'].keys())
        assert col_names == row_names
        matrix = pd.DataFrame(
            data=[data['data']['profiles'][r] for r in col_names], 
            columns=col_names, index=col_names
        )
        return matrix


class PrereleasedStreamlineCountQuery(FeatureQuery, PrereleasedConnectivityFetcher):

    _FEATURETYPE = StreamlineCounts

    def __init__(self, **kwargs):
        FeatureQuery.__init__(self)
        PrereleasedConnectivityFetcher.__init__(self, ["count", "1000brains"])
        for parc_id, matrixloader, srcinfo in self.get_data():
            self.register(self._FEATURETYPE(parc_id, matrixloader, srcinfo ))

class PrereleasedStreamlineLengthQuery(FeatureQuery, PrereleasedConnectivityFetcher):

    _FEATURETYPE = StreamlineLengths

    def __init__(self, **kwargs):
        FeatureQuery.__init__(self)
        PrereleasedConnectivityFetcher.__init__(self, ["length"])
        for parc_id, matrixloader, srcinfo in self.get_data():
            self.register(self._FEATURETYPE(parc_id, matrixloader, srcinfo ))

class PrereleasedRestingStateQuery(FeatureQuery, PrereleasedConnectivityFetcher):

    _FEATURETYPE = FunctionalConnectivity

    def __init__(self, **kwargs):
        FeatureQuery.__init__(self)
        keyword = "rest_fix"
        PrereleasedConnectivityFetcher.__init__(self, [keyword])
        for parc_id, matrixloader, srcinfo in self.get_data():
            filename = srcinfo['filename']
            paradigm = filename[filename.lower().find(keyword.lower()):]
            self.register(self._FEATURETYPE(parc_id, matrixloader, paradigm, srcinfo ))

#class ThousandBrainsStreamlineCountsQuery(FeatureQuery):
#
#    _FEATURETYPE = StreamlineCounts
#    _QUERY = GitlabConnector(
#        "https://jugit.fz-juelich.de", 3009, "develop"
#    ) 
#
#    def __init__(self, **kwargs):
#        FeatureQuery.__init__(self)
#        for _, loader in self._QUERY.get_loaders("connectivity", ".json"):
#            cm = ConnectivityMatrix._from_json(loader.data)
#            for parcellation in cm.parcellations:
#                species = [atlas.species for atlas in parcellation.atlases]
#                for regionname in cm.regionnames:
#                    region = parcellation.decode_region(regionname, build_group=False)
#                    if region is None:
#                        raise RuntimeError(
#                            f"Could not decode region name {regionname} in {parcellation}"
#                        )
#                    self.register(self._FEATURETYPE(parcellation, matrix))


# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .cache import CACHE
from ..commons import logger

import json
from zipfile import ZipFile
import requests
import os
from nibabel import Nifti1Image, GiftiImage, streamlines
import gzip
from getpass import getpass
from io import BytesIO
import urllib
import pandas as pd
from tempfile import mktemp


def bytes_to_temporary_zipfile(b):
    fname = mktemp(suffix=".zip")
    with open(fname, "wb") as f:
        f.write(b)
    return ZipFile(fname)


DECODERS = {
    ".nii.gz": lambda b: Nifti1Image.from_bytes(gzip.decompress(b)),
    ".nii": lambda b: Nifti1Image.from_bytes(b),
    ".gii": lambda b: GiftiImage.from_bytes(b),
    ".json": lambda b: json.loads(b.decode()),
    ".tck": lambda b: streamlines.load(BytesIO(b)),
    ".csv": lambda b: pd.read_csv(BytesIO(b), delimiter=";"),
    ".txt": lambda b: pd.read_csv(BytesIO(b), delimiter=" ", header=None),
    ".zip": bytes_to_temporary_zipfile,
}


class SiibraHttpRequestError(Exception):
    def __init__(self, response, msg="Cannot execute http request."):
        self.response = response
        self.msg = msg
        Exception.__init__(self)

    def __str__(self):
        return (
            f"{self.msg}\n"
            f"    Status code: {self.response.status_code}\n"
            f"    Url:         {self.response.url}\n"
        )


class HttpRequest:
    def __init__(
        self,
        url,
        func=None,
        msg_if_not_cached=None,
        refresh=False,
        post=False,
        **kwargs,
    ):
        """
        Initialize a cached http data loader.
        It takes a URL and optional data conversion function.
        For loading, the http request is only performed if the
        result is not yet available in the disk cache.
        Leaves the interpretation of the returned content to the caller.

        Parameters
        ----------
        url : string, or None
            URL for loading raw data, which is then fed into `func`
            for creating the output.
            If None, `func` will be called without arguments.
        func : function pointer
            Function for constructing the output data
            (called on the data retrieved from `url`, if supplied)
        refresh : bool, default: False
            If True, a possibly cached content will be ignored and refreshed
        post: bool, default: False
            perform a post instead of get
        """
        assert url is not None
        self.url = url
        suitable_decoders = [dec for sfx, dec in DECODERS.items() if url.endswith(sfx)]
        if (func is None) and (len(suitable_decoders) > 0):
            assert len(suitable_decoders) == 1
            self.func = suitable_decoders[0]
        else:
            self.func = func
        self.kwargs = kwargs
        self.cachefile = CACHE.build_filename(self.url + json.dumps(kwargs))
        self.msg_if_not_cached = msg_if_not_cached
        self.refresh = refresh
        self.post = post
        self._set_decoder_func(func, url)

    def _set_decoder_func(self, func, fileurl: str):
        urlpath = urllib.parse.urlsplit(fileurl).path
        if func is None:
            suitable_decoders = [
                dec for sfx, dec in DECODERS.items() if urlpath.endswith(sfx)
            ]
            if len(suitable_decoders) > 0:
                assert len(suitable_decoders) == 1
                self.func = suitable_decoders[0]
                return
        self.func = func

    @property
    def cached(self):
        return os.path.isfile(self.cachefile)

    def _retrieve(self):
        # Loads the data from http if required.
        # If the data is already cached, None is returned,
        # otherwise data (as it is already in memory anyway).
        # The caller should load the cachefile only
        # if None is returned.

        if self.cached and not self.refresh:
            # in cache. Just load the file
            logger.debug(
                f"Already in cache at {os.path.basename(self.cachefile)}: {self.url}"
            )
            return
        else:
            # not yet in cache, perform http request.
            logger.debug(f"Loading {self.url} to {os.path.basename(self.cachefile)}")
            if self.msg_if_not_cached is not None:
                logger.info(self.msg_if_not_cached)
            if self.post:
                r = requests.post(self.url, **self.kwargs)
            else:
                r = requests.get(self.url, **self.kwargs)
            if r.ok:
                with open(self.cachefile, "wb") as f:
                    f.write(r.content)
                self.refresh = False
                return r.content
            else:
                raise SiibraHttpRequestError(r)

    def get(self):
        data = self._retrieve()
        if data is None:
            with open(self.cachefile, "rb") as f:
                data = f.read()
        return data if self.func is None else self.func(data)

    @property
    def data(self):
        # for backward compatibility with old LazyHttpRequest class
        return self.get()


class ZipfileRequest(HttpRequest):
    def __init__(self, url, filename, func=None):
        HttpRequest.__init__(self, url, func=func)
        self.filename = filename
        self._set_decoder_func(func, filename)

    def get(self):
        self._retrieve()
        zipfile = ZipFile(self.cachefile)
        filenames = zipfile.namelist()
        matches = [fn for fn in filenames if fn.endswith(self.filename)]
        if len(matches) == 0:
            raise RuntimeError(
                f"Requested filename {self.filename} not found in archive at {self.url}"
            )
        if len(matches) > 1:
            raise RuntimeError(
                f'Requested filename {self.filename} was not unique in archive at {self.url}. Candidates were: {", ".join(matches)}'
            )
        with zipfile.open(matches[0]) as f:
            data = f.read()
        return data if self.func is None else self.func(data)


class EbrainsRequest(HttpRequest):
    """
    Implements lazy loading of HTTP Knowledge graph queries.
    """

    _KG_API_TOKEN = None
    keycloak_endpoint = (
        "https://iam.ebrains.eu/auth/realms/hbp/protocol/openid-connect/token"
    )

    def __init__(
        self, url, decoder=None, params={}, msg_if_not_cached=None, post=False
    ):
        """Construct an EBRAINS request."""
        # NOTE: we do not pass params and header here,
        # since we want to evaluate them late in the get() method.
        # This is nice because it allows to set env. variable KG_TOKEN only when
        # really needed, and not necessarily on package initialization.
        self.params = params
        HttpRequest.__init__(self, url, decoder, msg_if_not_cached, post=post)

    @classmethod
    def fetch_token(cls):
        """Fetch an EBRAINS token using commandline-supplied username/password
        using the data proxy endpoint.
        """
        username = input("Your EBRAINS username: ")
        password = getpass("Your EBRAINS password: ")
        response = requests.post(
            "https://data-proxy.ebrains.eu/api/auth/token",
            headers={
                "accept": "application/json",
                "Content-Type": "application/json",
            },
            data=f'{{"username": "{username}","password": "{password}"}}',
        )
        if response.status_code == 200:
            cls._KG_API_TOKEN = response.json()
        else:
            if response.status_code == 500:
                logger.error(
                    "Invalid EBRAINS username/password provided for fetching token."
                )
            raise SiibraHttpRequestError(response)

    @classmethod
    def set_token(cls, token):
        logger.info(f"Setting EBRAINS Knowledge Graph authentication token: {token}")
        cls._KG_API_TOKEN = token

    @property
    def kg_token(self):

        # token is available, return it
        if self.__class__._KG_API_TOKEN is not None:
            return self.__class__._KG_API_TOKEN

        # See if a token is directly provided in  $HBP_AUTH_TOKEN
        if "HBP_AUTH_TOKEN" in os.environ:
            self.__class__._KG_API_TOKEN = os.environ["HBP_AUTH_TOKEN"]
            return self.__class__._KG_API_TOKEN

        # try KEYCLOAK. Requires the following environment variables set:
        # KEYCLOAK_ENDPOINT, KEYCLOAK_CLIENT_ID, KEYCLOAK_CLIENT_SECRET
        keycloak = {
            v: os.environ.get(f"KEYCLOAK_{v.upper()}")
            for v in ["client_id", "client_secret"]
        }
        if None not in keycloak.values():
            logger.info("Getting an EBRAINS token via keycloak client configuration...")
            result = requests.post(
                self.__class__.keycloak_endpoint,
                data=(
                    f"grant_type=client_credentials&client_id={keycloak['client_id']}"
                    f"&client_secret={keycloak['client_secret']}"
                    "&scope=kg-nexus-role-mapping%20kg-nexus-service-account-mock"
                ),
                headers={"content-type": "application/x-www-form-urlencoded"},
            )
            try:
                content = json.loads(result.content.decode("utf-8"))
            except json.JSONDecodeError as error:
                logger.error(f"Invalid json from keycloak:{error}")
                self.__class__._KG_API_TOKEN = None
            if "error" in content:
                logger.error(content["error_description"])
                self.__class__._KG_API_TOKEN = None
            self.__class__._KG_API_TOKEN = content["access_token"]

        if self.__class__._KG_API_TOKEN is None:
            # No success getting the token
            raise RuntimeError(
                "No access token for EBRAINS Knowledge Graph found. "
                "If you do not have an EBRAINS account, please first register at "
                "https://ebrains.eu/register. Then, use one of the following option: "
                "\n 1. Let siibra get you a token by passing your username and password, using siibra.fetch_ebrains_token()"
                "\n 2. If you know how to get a token yourself, set it as $HBP_AUTH_TOKEN or siibra.set_ebrains_token()"
                "\n 3. If you are an application developer, you might configure keycloak access by setting $KEYCLOAK_ENDPOINT, "
                "$KEYCLOAK_CLIENT_ID and $KEYCLOAK_CLIENT_SECRET."
            )

        return self.__class__._KG_API_TOKEN

    @property
    def auth_headers(self):
        return {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.kg_token}",
        }

    def get(self):
        """Evaluate KG Token is evaluated only on executrion of the request."""
        self.kwargs = {"headers": self.auth_headers, "params": self.params}
        return super().get()


class EbrainsKgQuery(EbrainsRequest):
    """Request outputs from a knowledge graph query."""

    server = "https://kg.humanbrainproject.eu"
    org = "minds"
    domain = "core"
    version = "v1.0.0"

    SC_MESSAGES = {
        401: "The provided EBRAINS authentication token is not valid",
        403: "No permission to access the given query",
        404: "Query with this id not found",
    }

    def __init__(self, query_id, instance_id=None, schema="dataset", params={}):
        inst_tail = "/" + instance_id if instance_id is not None else ""
        self.schema = schema
        url = "{}/query/{}/{}/{}/{}/{}/instances{}?databaseScope=RELEASED".format(
            self.server,
            self.org,
            self.domain,
            self.schema,
            self.version,
            query_id,
            inst_tail,
        )
        EbrainsRequest.__init__(
            self,
            url,
            decoder=DECODERS[".json"],
            params=params,
            msg_if_not_cached=f"Executing EBRAINS KG query {query_id}{inst_tail}",
        )

    def get(self):
        try:
            result = EbrainsRequest.get(self)
        except SiibraHttpRequestError as e:
            if e.response.status_code in self.SC_MESSAGES:
                raise RuntimeError(self.SC_MESSAGES[e.response.status_code])
            else:
                raise RuntimeError(
                    f"Could not process HTTP request (status code: "
                    f"{e.response.status_code}). Message was: "
                    f"{e.response.message} "
                )
        return result



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import logger
from ..commons import MapType
from ..core.datasets import Dataset
from ..core.space import Space, BoundingBox
from ..retrieval import HttpRequest, ZipfileRequest, CACHE, SiibraHttpRequestError

from ctypes import ArgumentError
import numpy as np
import nibabel as nib
import os
import json
from abc import ABC, abstractmethod
from neuroglancer_scripts.precomputed_io import get_IO_for_existing_dataset
from neuroglancer_scripts.accessor import get_accessor_for_url
class VolumeSrc(Dataset, type_id="fzj/tmp/volume_type/v0.0.1"):

    _SPECIALISTS = {}
    volume_type = None
    _SURFACE_TYPES = ['gii']

    def __init__(self, identifier, name, url, space, detail=None, **kwargs):
        """
        Construct a new volume source.

        Parameters
        ----------
        identifier : str
            A unique identifier for the source
        name : str
            A human-readable name
        volume_type : str
            Type of volume source, clarifying the data format. Typical names: "nii", "neuroglancer/precomputed".
        url : str
            The URL to the volume src, typically a url to the corresponding image or tilesource.
        space : Space
            Reference space in which this volume is defined
        detail : dict
            Detailed information. Currently only used to store a transformation matrix  for neuroglancer tilesources.
        zipped_file : str
            The filename to be extracted from a zip file. If given, the url is
            expected to point to a downloadable zip archive. Currently used to
            extreact niftis from zip archives, as for example in case of the
            MNI reference templates.
        """
        Dataset.__init__(self, identifier=identifier)
        assert name is not None
        self.name = name
        self.url = url
        if "SIIBRA_URL_MODS" in os.environ and url:
            mods = json.loads(os.environ["SIIBRA_URL_MODS"])
            for old, new in mods.items():
                self.url = self.url.replace(old, new)
            if self.url != url:
                logger.warning(f"Applied URL modification\nfrom {url}\nto   {self.url}")
        self.detail = {} if detail is None else detail
        self.space = space
        self.map_type = None

    def __init_subclass__(cls, volume_type=None):
        """Called when this class gets subclassed by cls."""
        cls.volume_type = volume_type
        if volume_type is not None:
            assert volume_type not in VolumeSrc._SPECIALISTS
            VolumeSrc._SPECIALISTS[volume_type] = cls

    def __str__(self):
        return f"{self.volume_type} {self.url}"

    @property
    def is_volume(self):
        """Overwrite Dataset's default 'false'"""
        return True

    @property
    def is_surface(self):
        """Overwrite Dataset's default 'false'"""
        return self.volume_type in self._SURFACE_TYPES

    @classmethod
    def _from_json(cls, obj):
        """
        Provides an object hook for the json library to construct a VolumeSrc
        object from a json stream.
        """
        if obj.get("@type") != "fzj/tmp/volume_type/v0.0.1":
            raise NotImplementedError(
                f"Cannot build VolumeSrc from this json spec: {obj}"
            )

        volume_type = obj.get("volume_type")
        detail = obj.get("detail")
        url = obj.get("url")
        space = Space.REGISTRY[obj.get("space_id")]
        transform_nm = np.identity(4)
        if detail is not None and "neuroglancer/precomputed" in detail:
            if "transform" in detail["neuroglancer/precomputed"]:
                transform_nm = np.array(detail["neuroglancer/precomputed"]["transform"])

        # decide if object should be generated with a specialized derived class
        VolumeClass = cls._SPECIALISTS.get(volume_type, cls)
        kwargs = {
            "transform_nm": transform_nm,
            "zipped_file": obj.get("zipped_file", None),
        }
        if VolumeClass == cls:
            logger.error(f"Volume will be generated as plain VolumeSrc: {obj}")
        result = VolumeClass(
            identifier=obj["@id"],
            name=obj["name"],
            url=url,
            space=space,
            detail=detail,
            **kwargs,
        )
        # result.volume_type = obj.get('volume_type',None)
        maptype = obj.get("map_type", None)
        if maptype is not None:
            result.map_type = MapType[maptype.upper()]
        return result


class ImageProvider(ABC):
    @abstractmethod
    def fetch(self, resolution_mm=None, voi=None, mapindex=None, clip=False):
        """
        Provide access to image data.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not implement fetch(), use a derived class."
        )

    @abstractmethod
    def get_shape(self, resolution_mm=None):
        """
        Return the shape of the image volume.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not implement get_shape(), use a derived class."
        )

    @abstractmethod
    def is_float(self):
        """
        Return True if the data type of the volume is a float type.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not implement is_float(), use a derived class."
        )

    def is_4D(self):
        return len(self.get_shape()) == 4


class LocalNiftiVolume(ImageProvider):

    volume_type = 'nii'

    def __init__(self, name: str, img: nib.Nifti1Image, space: Space):
        """ Create a new local nifti volume from a Nifti1Image object.

        Args:
            name ([str]): A human-readable name
            img (Nifti1Image): 3D image
            space (Space): the reference space associated with the Image
        """
        self.name = name
        if isinstance(img, nib.Nifti1Image):
            self.image = img
        elif isinstance(img, str): 
            self.image = nib.load(img)
        else:
            raise ValueError(f"Cannot create local nifti volume from image parameter {img}") 

        if np.isnan(self.image.get_fdata()).any():
            logger.warn(f'Replacing NaN by 0 for {self.name}')
            self.image = nib.Nifti1Image(
                np.nan_to_num(self.image.get_fdata()),
                self.image.affine
            )
        self.space = space

    def fetch(self, **kwargs):
        return self.image

    def get_shape(self):
        return self.image.shape

    def is_float(self):
        return self.image.dataobj.dtype.kind == "f"

class RemoteNiftiVolume(ImageProvider, VolumeSrc, volume_type="nii"):

    _image_cached = None

    def __init__(
        self, identifier, name, url, space, detail=None, zipped_file=None, **kwargs
    ):
        VolumeSrc.__init__(self, identifier, name, url, space, detail=detail)
        if zipped_file is None:
            self._image_loader = HttpRequest(url)
        else:
            self._image_loader = ZipfileRequest(url, zipped_file)

    @property
    def image(self):
        return self._image_loader.data

    def fetch(self, resolution_mm=None, voi=None, mapindex=None, clip=False):
        """
        Loads and returns a Nifti1Image object representing the volume source.

        Parameters
        ----------
        resolution_mm : float or None (Default: None)
            Request the template at a particular physical resolution in mm. If None,
            the native resolution is used.
            Currently, this only works for neuroglancer volumes.
        voi : BoundingBox
            optional bounding box
        clip : Boolean, default: False
            if True, generates an object where the image data array is cropped to its bounding box (with properly adjusted affine matrix)
        """
        if clip and voi:
            raise ArgumentError("voi and clip cannot only be requested independently")
        shape = self.get_shape()
        img = None
        if resolution_mm is not None:
            raise NotImplementedError(
                f"NiftiVolume does not support to specify image resolutions (but {resolution_mm} was given)"
            )

        if mapindex is None:
            img = self.image
        elif len(shape) != 4 or mapindex >= shape[3]:
            raise IndexError(
                f"Mapindex of {mapindex} provided for fetching from NiftiVolume, but its shape is {shape}."
            )
        else:
            img = nib.Nifti1Image(
                dataobj=self.image.dataobj[:, :, :, mapindex], affine=self.image.affine
            )
        assert img is not None

        bb_vox = None
        if voi is not None:
            bb_vox = voi.transform_bbox(np.linalg.inv(img.affine))
        elif clip:
            # determine bounding box by cropping the nonzero values
            bb_vox = BoundingBox.from_image(img)

        if bb_vox is not None:
            (x0, y0, z0), (x1, y1, z1) = bb_vox.minpoint, bb_vox.maxpoint
            shift = np.identity(4)
            shift[:3, -1] = bb_vox.minpoint
            img = nib.Nifti1Image(
                dataobj=img.dataobj[x0:x1, y0:y1, z0:z1],
                affine=np.dot(img.affine, shift),
            )

        return img

    def get_shape(self, resolution_mm=None):
        if resolution_mm is not None:
            raise NotImplementedError(
                "NiftiVolume does not support to specify different image resolutions"
            )
        try:
            return self.image.shape
        except AttributeError as e:
            logger.error(
                f"Invalid object type {type(self.image)} of image for {self} {self.name}"
            )
            raise (e)

    def is_float(self):
        return self.image.dataobj.dtype.kind == "f"


class NeuroglancerVolume(
    ImageProvider, VolumeSrc, volume_type="neuroglancer/precomputed"
):
    # Number of bytes at which an image array is considered to large to fetch
    MAX_GiB = .2

    @property
    def MAX_BYTES(self):
        return self.MAX_GiB * 1024**3
    
    def __init__(self, identifier, name, url, space, detail, **kwargs):

        VolumeSrc.__init__(self, identifier, name, url, space, detail)
        ImageProvider.__init__(self)
        
        self.url = url
        self.space = space
        
        accessor = get_accessor_for_url(url)
        self._io = get_IO_for_existing_dataset(accessor)
        self.scales = {}
        self.dtype = np.dtype(self._io.info['data_type'])
        self.scales = sorted([NeuroglancerScale(self, i) for i in self._io.info['scales']])

        try:
            res = HttpRequest(f"{self.url}/transform.json").get()
        except SiibraHttpRequestError:
            res = None
        if res is not None:
            logger.debug("Found global affine transform file, intrepreted in nanometer space.")
            self.transform_nm = np.array(res)
        else:
            self.transform_nm = np.identity(4)
            
    def fetch(self, resolution_mm = None, voi: BoundingBox = None):
        if voi is not None:
            assert voi.space == self.space
        scale = self._select_scale(resolution_mm, voi)
        logger.info(
            f"Fetching resolution "
            f"{', '.join(map('{:.2f}'.format, scale.res_mm))} mm "
            f"to provide resolution {resolution_mm}."
        )
        return scale.fetch(voi)

    def get_shape(self, resolution_mm=None):
        scale = self._select_scale(resolution_mm)
        return scale.size 

    def is_float(self):
        return self.dtype.kind == "f"
        
    def _select_scale(self, resolution_mm, bbox: BoundingBox = None ):
        if resolution_mm is None:
            suitable = self.scales
        else:
            suitable = sorted(s for s in self.scales if s.resolves(resolution_mm))
        if len(suitable)>0:
            scale = suitable[-1]
        else:
            scale = self.scales[0]
            logger.warn(
                f"Requested resolution {resolution_mm} is not available. "
                f"Falling back to the highest possible resolution of "
                f"{', '.join(map('{:.2f}'.format, scale.res_mm))} mm."
            )
        while scale._estimate_nbytes(bbox) > self.MAX_BYTES:
            scale = scale.next()
            if scale is None:
                raise RuntimeError(
                    f"Fetching bounding box {bbox} is infeasible "
                    f"relative to the limit of {self.MAX_BYTES/1024**3}GiB."
                )
        return scale
    

class NeuroglancerScale:
    """One scale of a NeuroglancerVolume.
    """
    
    def __init__(self, volume: NeuroglancerVolume, scaleinfo: dict):
        self.volume = volume
        self.chunk_sizes = np.array(scaleinfo['chunk_sizes']).squeeze()
        self.encoding = scaleinfo['encoding']
        self.key = scaleinfo['key']
        self.res_nm = np.array(scaleinfo['resolution']).squeeze()
        self.size = scaleinfo['size']
        self.voxel_offset = np.array(scaleinfo['voxel_offset'])
        
    @property
    def res_mm(self):
        return self.res_nm/1e6
    
    @property
    def space(self):
        """ forward the corresponding volume's coordinate space. """
        return self.volume.space
        
    def resolves(self, resolution_mm):
        """ Test wether the resolution of this scale is sufficient to provide the given resolution. """
        return any(r/1e6<=resolution_mm for r in self.res_nm)
        
    def __lt__(self, other):
        """ Sort scales by resolution. """
        return all(self.res_nm[i] < other.res_nm[i] for i in range(3))
    
    def __repr__(self):
        return str(self)
    
    def __str__(self):
        return f"{self.__class__.__name__} {self.key}"
    
    def _estimate_nbytes(self, bbox: BoundingBox= None):
        """ Estimate the size image array to be fetched in bytes, given a bounding box."""
        if bbox is None:
            bbox_ = BoundingBox(
                (0, 0, 0), self.size, self.space
            )
        else:
            bbox_ = bbox.transform(np.linalg.inv(self.affine))
        result = self.volume.dtype.itemsize * bbox_.volume
        logger.info(
            f"Approximate size for fetching resolution "
            f"({', '.join(map('{:.2f}'.format, self.res_mm))}) mm "
            f"is {result/1024**3:.2f} GiB."
        )
        return result
    
    def next(self):
        """ Returns the next scale in this volume, of None if this is the last.
        """
        my_index = self.volume.scales.index(self)
        print(f"Index of {self.key} is {my_index} of {len(self.volume.scales)}.")
        if my_index < len(self.volume.scales):
            return self.volume.scales[my_index+1]
        else:
            return None
            
    def prev(self):
        """ Returns the previous scale in this volume, or None if this is the first. 
        """
        my_index = self.volume.scales.index(self)
        print(f"Index of {self.key} is {my_index} of {len(self.volume.scales)}.")
        if my_index > 0:
            return self.volume.scales[my_index-1]
        else:
            return None

    @property
    def affine(self):
        scaling = np.diag(np.r_[self.res_nm, 1.])
        affine = np.dot(self.volume.transform_nm, scaling)
        affine[:3, :] /= 1e6
        return affine
    
    def _chunk_of_point(self, xyz):
        return np.floor((np.array(xyz) - self.voxel_offset) / self.chunk_sizes).astype('int').ravel()
    
    def _read_chunk(self, gx, gy, gz):
        cachefile = CACHE.build_filename(
            "{}_{}_{}_{}_{}".format(self.volume.url, self.key, gx, gy, gz),
            suffix='.npy'
        )
        if os.path.isfile(cachefile):
            return np.load(cachefile)
                
        x0 = gx * self.chunk_sizes[0]
        y0 = gy * self.chunk_sizes[1]
        z0 = gz * self.chunk_sizes[2]
        x1, y1, z1 = np.minimum(self.chunk_sizes + [x0, y0, z0], self.size)
        chunk_czyx = self.volume._io.read_chunk(
            self.key, (x0, x1, y0, y1, z0, z1)
        )
        if not chunk_czyx.shape[0] == 1:
            raise NotImplementedError("Color channel data is not yet supported")
        chunk_zyx = chunk_czyx[0]
        np.save(cachefile, chunk_zyx)
        return chunk_zyx
    
    def fetch(self, voi: BoundingBox = None):
        
        # define the bounding box in this scale's voxel space
        if voi is None:
            bbox_ = BoundingBox(
                (0, 0, 0), self.size, self.space
            )
        else:
            bbox_ = voi.transform(np.linalg.inv(self.affine))        
        
        # extract minimum and maximum the chunk indices to be loaded 
        gx0, gy0, gz0 = self._chunk_of_point(tuple(bbox_.minpoint))
        gx1, gy1, gz1 = self._chunk_of_point(tuple(bbox_.maxpoint))+1

        # create requested data volume, and fill it with the required chunk data
        shape_zyx = (np.array([gz1-gz0, gy1-gy0, gx1-gx0]) * self.chunk_sizes[::-1])
        data_zyx = np.zeros(shape_zyx, dtype=self.volume.dtype)
        for gx in range(gx0, gx1):
            x0 = (gx-gx0)*self.chunk_sizes[0]
            for gy in range(gy0, gy1):
                y0 = (gy-gy0)*self.chunk_sizes[1]
                for gz in range(gz0, gz1):
                    z0 = (gz-gz0)*self.chunk_sizes[2]
                    chunk = self._read_chunk(gx, gy, gz)
                    z1, y1, x1 = np.array([z0, y0, x0]) + chunk.shape
                    data_zyx[z0:z1, y0:y1, x0:x1] = chunk
                    
        # determine the remaining offset from the "chunk mosaic" to the
        # exact bounding box requested, to cut off undesired borders
        data_min = np.array([gx0, gy0, gz0]) * self.chunk_sizes
        x0, y0, z0 = (np.array(tuple(bbox_.minpoint)) - data_min).astype('int')
        xd, yd, zd = np.array(bbox_.shape).astype('int')
        offset = tuple(bbox_.minpoint)
                
        # build the nifti image
        trans = np.identity(4)[[2,1,0,3], :] # zyx -> xyz
        shift = np.c_[np.identity(4)[:,:3], np.r_[offset, 1]]
        return nib.Nifti1Image(
            data_zyx[z0:z0+zd, y0:y0+yd, x0:x0+xd], 
            np.dot(self.affine, np.dot(shift, trans))
        )




# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .feature import SpatialFeature
from .query import FeatureQuery

from .. import QUIET
from ..volumes.volume import VolumeSrc
from ..core.space import Space, BoundingBox
from ..core.datasets import EbrainsDataset
from ..retrieval.repositories import GitlabConnector

import numpy as np


class VolumeOfInterest(SpatialFeature, EbrainsDataset):
    def __init__(self, dataset_id, location, name):
        SpatialFeature.__init__(self, location)
        EbrainsDataset.__init__(self, dataset_id, name)
        self.volumes = []

    @classmethod
    def _from_json(cls, definition):
        if definition["@type"] == "minds/core/dataset/v1.0.0":
            space = Space.REGISTRY[definition["space id"]]
            vsrcs = []
            minpoints = []
            maxpoints = []
            for vsrc_def in definition["volumeSrc"]:
                vsrc = VolumeSrc._from_json(vsrc_def)
                vsrc.space = space
                with QUIET:
                    img = vsrc.fetch()
                    D = np.asanyarray(img.dataobj).squeeze()
                    nonzero = np.array(np.where(D > 0))
                    A = img.affine
                minpoints.append(np.dot(A, np.r_[nonzero.min(1)[:3], 1])[:3])
                maxpoints.append(np.dot(A, np.r_[nonzero.max(1)[:3], 1])[:3])
                vsrcs.append(vsrc)
            minpoint = np.array(minpoints).min(0)
            maxpoint = np.array(maxpoints).max(0)
            result = cls(
                dataset_id=definition["kgId"],
                name=definition["name"],
                location=BoundingBox(minpoint, maxpoint, space),
            )
            list(map(result.volumes.append, vsrcs))
            return result
        return definition


class VolumeOfInterestQuery(FeatureQuery):
    _FEATURETYPE = VolumeOfInterest
    _QUERY = GitlabConnector("https://jugit.fz-juelich.de", 3009, "develop")

    def __init__(self, **kwargs):
        FeatureQuery.__init__(self)
        for _, loader in self._QUERY.get_loaders(folder="vois", suffix=".json"):
            voi = VolumeOfInterest._from_json(loader.data)  # json.loads(data))
            self.register(voi)



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .volume import VolumeSrc

from ..retrieval import HttpRequest

import numpy as np

class GiftiSurface(VolumeSrc, volume_type="gii"):
    """
    A (set of) surface meshes in Gifti format.
    """
    def __init__(self, identifier, name, url, space, detail=None, **kwargs):
        VolumeSrc.__init__(self, identifier, name, url, space, detail, **kwargs)
        if isinstance(url, str):
            # a single url
            self._loaders = {name: url}
        elif isinstance(url, dict):
            self._loaders = {
                label: HttpRequest(url)
                for label, url in url.items()
            } 
        else:
            raise NotImplementedError(f"Urls for {self.__class__.__name__} are expected to be of type str or dict.")

    def fetch(self, name=None):
        """
        Returns the mesh as a dictionary with two numpy arrays: An Nx3 array of vertex coordinates, 
        and an Mx3 array of face definitions using row indices of the vertex array.

        If name is specified, only submeshes matching this name are included, otherwise all meshes are combined.
        """
        vertices = np.empty((0, 3))
        faces = np.empty((0,3), dtype='int')
        for n, loader in self._loaders.items():
            npoints = vertices.shape[0]
            if (name is not None) & (n!=name):
                continue
            assert len(loader.data.darrays)>1
            vertices = np.append(vertices, loader.data.darrays[0].data, axis = 0)
            faces = np.append(faces, loader.data.darrays[1].data+npoints, axis = 0)

        return dict(zip(['verts','faces'], [vertices, faces]))

    @property
    def variants(self):
        return list(self._loaders.keys())

    def fetch_iter(self):
        """
        Iterator returning all submeshes, each as a tuple consisting of 
        - their name (string)
        - the mesh, as a dictionary of two numpy arrays: 
          An Nx3 array of vertex coordinates, and an Mx3 array 
          of face definitions using row indices of the vertex array.
        """
        return ((v,self.fetch(v)) for v in self.variants)

class GiftiSurfaceLabeling(VolumeSrc, volume_type="gii-label"):
    """
    A mesh labeling, specified by a gifti file.
    """
    def __init__(self, identifier, name, url, space, detail=None, **kwargs):
        VolumeSrc.__init__(self, identifier, name, url, space, detail, **kwargs)
        self._loader = HttpRequest(self.url)

    def fetch(self):
        """Returns a 1D numpy array of label indices."""
        assert(len(self._loader.data.darrays)==1)
        return self._loader.data.darrays[0].data


class NeuroglancerMesh(VolumeSrc, volume_type="neuroglancer/precompmesh"):
    """
    A surface mesh provided as neuroglancer precomputed mesh.     
    """

    def __init__(self, identifier, name, url, space, detail=None, **kwargs):
        VolumeSrc.__init__(self, identifier, name, url, space, detail, **kwargs)

    def fetch(self):
        raise NotImplementedError(f"Fetching from neuroglancer precomputed mesh is not yet implemented.")



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .feature import Feature

from .. import logger, QUIET
from ..commons import Registry
from ..core import AtlasConcept, Dataset

from abc import ABC
from collections import defaultdict


class FeatureQuery(ABC):
    """
    An abstract class for data features extractors.
    """

    _FEATURETYPE = Feature
    _instances = {}
    _implementations = defaultdict(list)

    def __init__(self, **kwargs):
        logger.debug(f"Initializing query for {self._FEATURETYPE.__name__} features")
        self.features = []

    def __init_subclass__(cls):
        """
        Registers all subclasses of FeatureQuery.
        """
        logger.debug(
            f"New query {cls.__name__} for {cls._FEATURETYPE.__name__} features"
        )
        cls._implementations[cls._FEATURETYPE.modality()].append(cls)

    @classmethod
    def get_features(cls, concept, modality, group_by=None, **kwargs):
        """
        Retrieve data features linked to an atlas concept, by modality.

        Parameters
        ----------
        concept: AtlasConcept
            An atlas concept, typically a Parcellation or Region object.
        modality: FeatureType, list of FeatureType, or 'all'
            See siibra.features.modalities for available modalities.
            if 'all', all feature modalities will be queried.
        group_by: string, default: None
            If "dataset", group features additionally by dataset id.
            If "concept", group features additionally by the exact linked concept.
            Use the latter for example when you query by a parcellation, and want to get
            features grouped by the particular region of that parcellation which they have been attached to.
        """
        # helper func: convert modality spec into unique modality 
        mod = lambda modality: cls.get_modalities()[modality]

        if isinstance(modality, str) and modality == 'all':
            # use union of all featurequery implementations
            querytypes = sum(FeatureQuery._implementations.values(), [])
        elif isinstance(modality, (list, tuple)):
            # use union of featurequery implementations from multiple modalities
            querytypes = sum(
                [FeatureQuery._implementations[mod(m)] for m in modality], []
            )
        else:
            if mod(modality) not in cls._implementations:
                raise RuntimeError(
                    f"No feature query known for feature type {modality}."
                )
            querytypes = cls._implementations[mod(modality)]

        result = {}
        args_hash = hash(tuple(sorted(kwargs.items())))
        for querytype in querytypes:

            if (querytype, args_hash) not in cls._instances:
                logger.debug(f"Building new query {querytype} with args {kwargs}")
                try:
                    cls._instances[querytype, args_hash] = querytype(**kwargs)
                except TypeError as e:
                    logger.error(f"Cannot initialize {querytype} query: {str(e)}")
                    raise (e)

            query = cls._instances[querytype, args_hash]
            matches = query.execute(concept)
            logger.debug(f"{len(matches)} matches from query {query.__class__.__name__}")

            if group_by is None:
                if querytype.modality() not in result:
                    result[querytype.modality()] = []
                result[querytype.modality()].extend(matches)

            elif group_by == "dataset":
                if querytype.modality() not in result:
                    result[querytype.modality()] = defaultdict(list)
                for m in matches:
                    idf = m.id if isinstance(m, Dataset) else None
                    result[querytype.modality()][idf].append(m)

            elif group_by == "concept":
                if querytype.modality() not in result:
                    result[querytype.modality()] = defaultdict(list)
                for m in matches:
                    result[querytype.modality()][m._match].append(m)

            else:
                raise ValueError(
                    f"Invalid parameter '{group_by}' for the 'group_by' attribute of get_features. "
                    "Valid entries are: 'dataset', 'concept', or None.")


        # If only one modality was requested, simplify the dictionary
        if len(result) == 1:
            return next(iter(result.values()))
        else:
            return result

    @classmethod
    def get_modalities(cls):
        return Registry(
            elements={
                c:c for c in cls._implementations.keys()
            }
        )

    @classmethod
    def queries(cls, modality: str, **kwargs):
        """
        return global query objects for the given modality, remembering unique instances
        to avoid redundant FeatureQuery instantiations.
        """
        instances = []
        args_hash = hash(tuple(sorted(kwargs.items())))
        matched_modality = cls.get_modalities()[modality]
        for Querytype in cls._implementations[matched_modality]:
            if (Querytype, args_hash) not in cls._instances:
                logger.debug(f"Building new query {Querytype} with args {kwargs}")
                try:
                    cls._instances[Querytype, args_hash] = Querytype(**kwargs)
                except TypeError as e:
                    logger.error(f"Cannot initialize {Querytype} query: {str(e)}")
                    raise (e)
            instances.append(cls._instances[Querytype, args_hash])
        return instances

    def execute(self, concept: AtlasConcept):
        """
        Executes a query for features associated with an atlas object.
        """
        matches = []
        for feature in self.features:
            if feature.match(concept):
                matches.append(feature)        
        return matches

    def __str__(self):
        return "\n".join([str(f) for f in self.features])

    def register(self, feature):
        assert isinstance(feature, self._FEATURETYPE)
        self.features.append(feature)

    @classmethod
    def modality(cls):
        return cls._FEATURETYPE.modality()
